from itertools import chain

import reframe as rfm
import reframe.utility.sanity as sn
from reframe.core.builtins import parameter, run_after, sanity_function, performance_function

from eessi.testsuite.constants import DEVICE_TYPES, COMPUTE_UNIT, CPU, NUMA_NODE, GPU
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.utils import find_modules


class EESSI_PyTorch_torchvision(rfm.RunOnlyRegressionTest, EESSI_Mixin):
    descr = 'Benchmark that runs a selected torchvision model on synthetic data'

    nn_model = parameter(['vgg16', 'resnet50', 'resnet152', 'densenet121', 'mobilenet_v3_large'])
    bench_name_ci = 'resnet50'
    parallel_strategy = parameter([None, 'ddp'])
    # Both torchvision and PyTorch-bundle modules have everything needed to run this test
    module_name = parameter(chain(find_modules('torchvision'), find_modules('PyTorch-bundle')))
    executable = 'python'
    time_limit = '30m'
    readonly_files = ['get_free_socket.py', 'pytorch_synthetic_benchmark.py']

    def required_mem_per_node(self):
        return self.num_tasks_per_node * 1024

    @run_after('init')
    def prepare_test(self):
        # Set nn_model as executable option
        self.executable_opts = ['pytorch_synthetic_benchmark.py --model %s' % self.nn_model]
        self.bench_name = self.nn_model

        # If not a GPU run, disable CUDA
        if self.device_type != DEVICE_TYPES[GPU]:
            self.executable_opts += ['--no-cuda']

    @run_after('setup')
    def set_ddp_options(self):
        "Set environment variables for PyTorch DDP"
        if self.parallel_strategy == 'ddp':
            # Set additional options required by DDP
            self.executable_opts += ["--master-port $(python get_free_socket.py)"]
            self.executable_opts += ["--master-address $(hostname --fqdn)"]
            self.executable_opts += ["--world-size %s" % self.num_tasks]

    @run_after('setup')
    def filter_invalid_parameter_combinations(self):
        # We cannot detect this situation before the setup phase, because it requires self.num_tasks.
        # Thus, the core count of the node needs to be known, which is only the case after the setup phase.
        msg = "Skipping test: parallel strategy is 'None',"
        msg += f" but requested process count is larger than one ({self.num_tasks})."
        self.skip_if(self.num_tasks > 1 and self.parallel_strategy is None, msg)
        msg = f"Skipping test: parallel strategy is {self.parallel_strategy},"
        msg += " but only one process is requested."
        self.skip_if(self.num_tasks == 1 and self.parallel_strategy is not None, msg)

    @run_after('setup')
    def pass_parallel_strategy(self):
        "Set parallelization strategy when using more than one process"
        if self.num_tasks != 1:
            self.executable_opts += ['--use-%s' % self.parallel_strategy]

    @sanity_function
    def assert_num_ranks(self):
        '''Assert that the number of reported CPUs/GPUs used is correct'''
        return sn.assert_found(r'Total img/sec on %s .PU\(s\):.*' % self.num_tasks, self.stdout)

    @performance_function('img/sec')
    def total_throughput(self):
        '''Total training throughput, aggregated over all CPUs/GPUs'''
        return sn.extractsingle(r'Total img/sec on [0-9]+ .PU\(s\):\s+(?P<perf>\S+)', self.stdout, 'perf', float)

    @performance_function('img/sec')
    def througput_per_CPU(self):
        '''Training througput per device type'''
        if self.device_type == DEVICE_TYPES[CPU]:
            return sn.extractsingle(r'Img/sec per CPU:\s+(?P<perf_per_cpu>\S+)', self.stdout, 'perf_per_cpu', float)
        else:
            return sn.extractsingle(r'Img/sec per GPU:\s+(?P<perf_per_gpu>\S+)', self.stdout, 'perf_per_gpu', float)


@rfm.simple_test
class EESSI_PyTorch_torchvision_CPU(EESSI_PyTorch_torchvision):
    device_type = DEVICE_TYPES[CPU]
    compute_unit = COMPUTE_UNIT[NUMA_NODE]


@rfm.simple_test
class EESSI_PyTorch_torchvision_GPU(EESSI_PyTorch_torchvision):
    device_type = DEVICE_TYPES[GPU]
    compute_unit = COMPUTE_UNIT[GPU]
    precision = parameter(['default', 'mixed'])

    @run_after('init')
    def prepare_gpu_test(self):
        # Set precision
        if self.precision == 'mixed':
            self.executable_opts += ['--use-amp']



# Author: Caspar van Leeuwen (SURF)
# Based on the example for the MultiWorkerMirroredStrategy with keras from
# https://www.tensorflow.org/tutorials/distribute/multi_worker_with_keras
import argparse
import json
import logging
import os
import sys
import socket

from contextlib import closing
from mpi4py import MPI
from timeit import default_timer as timer

import tensorflow as tf


def print0(msg, comm=MPI.COMM_WORLD):
    '''Prints string "msg" from rank 0'''
    output = comm.gather(msg, root=0)
    rank = comm.Get_rank()
    # Print elements per rank
    if rank == 0:
        for (rank, rank_out) in enumerate(output):
            print(f'Rank {rank}: {rank_out}')


def find_free_port():
    '''Function that gets a free port for the current process'''
    with closing(socket.socket()) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def get_local_rank(rank_info, rank_info_list):
    '''Function that figures out the local rank based on a list of rank,
    hostname, and port gathered from each of the workers'''
    # Note that rank_info_list is sorted by rank, by definition of the MPI allgather routine.
    # Thus, if our current rank is the n-th item in rank_info_list for which the hostname matches,
    # our local rank is n
    for index, item in enumerate(rank_info_list):
        if item['hostname'] == rank_info['hostname'] and item['rank'] == rank_info['rank']:
            return index


def get_rank_info(comm=MPI.COMM_WORLD):
    '''Create a dict for this worker containing rank, hostname and port to be used by this worker'''
    rank = comm.Get_rank()
    hostname = socket.gethostname()
    port = find_free_port()

    return {
        'rank': rank,
        'hostname': hostname,
        'port': port,
    }


def set_tf_config(rank_info, rank_info_list):
    '''Sets the TF_CONFIG environment variable for the current worker, based on the rank_info_list'''
    worker_list = ['%s:%s' % (item['hostname'], item['port']) for item in rank_info_list]

    tf_config = {
        'cluster': {
            'worker': worker_list,
        },
        'task': {'type': 'worker', 'index': rank_info['rank']}
    }
    os.environ["TF_CONFIG"] = json.dumps(tf_config)

    # logger.info(f"Set TF_CONFIG for rank {rank_info['rank']} to {tf_config}.")
    print0(f"Set TF_CONFIG for rank {rank_info['rank']} to {tf_config}.")
    return tf_config


parser = argparse.ArgumentParser(
    prog='Tensorflow Distributed Test',
    description='This program runs a distributed TensorFlow test using the tf.distribute.MultiWorkerMirroredStrategy'
                ' and the Keras fit API')

parser.add_argument(
    '-d', '--device', type=str, default='cpu', choices=['cpu', 'gpu'], help='Device to use for training')
parser.add_argument(
    '--inter-op-parallelism', type=int, default=1, help='Sets tf.config.threading.set_inter_op_parallelism_threads')
parser.add_argument(
    '--intra-op-parallelism', type=int, default=0, help='Sets tf.config.threading.set_intra_op_parallelism_threads')
parser.add_argument(
    '--per-worker-batch-size', type=int, default=4096, help='Batch size processed by each worker')
parser.add_argument(
    '--per-worker-test-batch-size', type=int, default=512,
    help='Batch size for computing accuracy on the validation set')
parser.add_argument(
    '--epochs-to-train', type=int, default=4, help='Number of epochs to train')
parser.add_argument(
    '--steps-per-epoch', type=int, default=25, help='Number of steps to train per epoch')
args = parser.parse_args()

# Make sure we can import mnist_setup from current dir
if '.' not in sys.path:
    sys.path.insert(0, '.')
import mnist_setup  # noqa: E402

os.environ.pop('TF_CONFIG', None)

# Set number of threads to use. Needs to be set early, before initialization
tf.config.threading.set_inter_op_parallelism_threads(args.inter_op_parallelism)
tf.config.threading.set_intra_op_parallelism_threads(args.intra_op_parallelism)

# Multi-worker config
# We'll use mpi4py to figure out our rank, have each process select a socket and hostname,
# and allreduce that information to create a TF_CONFIG
comm = MPI.COMM_WORLD
rank_info = get_rank_info(comm)
rank_info_list = comm.allgather(rank_info)
local_rank = get_local_rank(rank_info, rank_info_list)

# Create logger per rank
# logging.basicConfig(
#     filename='rank_%s.out' % local_rank,
#     format='%(asctime)s %(levelname)s %(message)s',
#     datefmt='%Y-%m-%d %H:%M:%S'
# )
# logging.info(f"Rank {rank_info['rank']} has local_rank {local_rank}, hostname {rank_info['hostname']}"
#              " and port {rank_info['port']}")
print0(f"Rank {rank_info['rank']} has local_rank {local_rank}, hostname {rank_info['hostname']}"
       " and port {rank_info['port']}")

# Turn off tensorflow info and warnings for rank != 0
if local_rank != 0:
    print("Turning off logging")
    # tf.get_logger().setLevel('ERROR')
    # tf.autograph.set_verbosity(1)
    # tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

tf_config = set_tf_config(rank_info, rank_info_list)
num_workers = len(tf_config['cluster']['worker'])

# Set visible devices and create MultiWorkerMirroredStrategy
# logging.info(f"Selecting device: {args.device}")
print0(f"Selecting device: {args.device}")
if args.device == 'gpu':
    # Limit each local rank to its own GPU.
    physical_devices = tf.config.list_physical_devices('GPU')
    try:
        # We could do local_rank % len(physical_devices) if we wanted to support running more than one rank per device
        # The current code doesn't support that
        tf.config.set_visible_devices(physical_devices[local_rank], 'GPU')
        visible_devices = tf.config.get_visible_devices('GPU')
        # logging.info("Local rank: %s, visible_devices: %s" % (local_rank, visible_devices))
        print0("Local rank: %s, visible_devices: %s" % (local_rank, visible_devices))
        assert len(visible_devices) == 1
    except Exception:
        print0("ERROR: Selection of GPU device based on local rank failed. Local rank: %s. Selected devices: %s"
               % (local_rank, visible_devices))
        # logging.error("Selection of GPU device based on local rank failed. Local rank: %s. Selected devices: %s"
        #       % (local_rank, visible_devices))

    # Should now have 1 GPU per process. Set memory growth for that device to avoid issues similar to
    # https://github.com/tensorflow/tensorflow/issues/45044
    # TODO: I have the feeling this is only needed because rank 0 somehow erroneously starts on CPU?
    # The fact that rank 0 started on CPU was triggered by setting OMP_PROC_BIND. I need to test if the following
    # line is still needed if we don't set OMP* binding variables
    tf.config.experimental.set_memory_growth(physical_devices[local_rank], True)

    # Set communication to NCCL explicitely
    communication_options = tf.distribute.experimental.CommunicationOptions(
        implementation=tf.distribute.experimental.CommunicationImplementation.NCCL)
    strategy = tf.distribute.MultiWorkerMirroredStrategy(communication_options=communication_options)
elif args.device == 'cpu':
    # Run on CPU, so make sure no GPU devices are visible
    tf.config.set_visible_devices([], 'GPU')
    visible_devices = tf.config.get_visible_devices()
    print0("Local rank: %s, visible_devices: %s" % (local_rank, visible_devices))
    # logging.info("Local rank: %s, visible_devices: %s" % (local_rank, visible_devices))
    strategy = tf.distribute.MultiWorkerMirroredStrategy()
# logging.info("Multiworker strategy created")
print0("Multiworker strategy created")

# Get datasets
global_batch_size = args.per_worker_batch_size * num_workers
global_test_batch_size = args.per_worker_test_batch_size * num_workers
multi_worker_dataset, multi_worker_test_dataset = mnist_setup.mnist_dataset(global_batch_size, global_test_batch_size)

with strategy.scope():
    multi_worker_model = mnist_setup.build_and_compile_cnn_model()

# Run the training
starttime = timer()
multi_worker_model.fit(
    multi_worker_dataset, epochs=args.epochs_to_train, steps_per_epoch=args.steps_per_epoch, verbose=2)
endtime = timer()
# logging.info("Keras fit completed!")
print0("Keras fit completed!")

# Compute performance
training_time = endtime - starttime
total_samples_trained = global_batch_size * args.steps_per_epoch * args.epochs_to_train
throughput = total_samples_trained / training_time
if local_rank == 0:
    print(f"Total training time: {training_time}")
    print(f"Performance: {throughput} img/s")

# Run evaluation to check accuracy on validation set
logging.info("Run evaluation")
metrics = multi_worker_model.evaluate(multi_worker_test_dataset, verbose=2)
if local_rank == 0:
    print(f"Final loss: {metrics[multi_worker_model.metrics_names.index('loss')]}")
    print(f"Final accuracy: {metrics[multi_worker_model.metrics_names.index('accuracy')]}")



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository
# 3. Temporarily change the 'access' field for the GPU partition to
# 'access':  ['-p gpu', '--export=None', '--gres=gpu:1'],

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job
# 3. Vega doesn't allow submission to the GPU partition without requesting at least one GPU (change #2)

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914

import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import *  # noqa: F403

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# This is an example configuration file
site_configuration = {
    'systems': [
        {
            'name': 'vega',
            'descr': 'Vega, a EuroHPC JU system',
            'modules_system': 'lmod',
            'hostnames': ['vglogin*', 'cn*', 'gn*'],
            'prefix': reframe_prefix,
            'partitions': [
                {
                    'name': 'cpu',
                    'scheduler': 'slurm',
                    'prepare_cmds': [
                        common_eessi_init(),
                        # Pass job environment variables like $PATH, etc., into job steps
                        'export SLURM_EXPORT_ENV=ALL',
                        # Needed when using srun launcher
                        # 'export SLURM_MPI_TYPE=pmix',  # WARNING: this broke the GROMACS on Vega
                        # Avoid https://github.com/EESSI/software-layer/issues/136
                        # Can be taken out once we don't care about old OpenMPI versions anymore (pre-4.1.1)
                        'export OMPI_MCA_pml=ucx',
                    ],
                    'launcher': 'mpirun',
                    # Use --export=None to avoid that login environment is passed down to submitted jobs
                    'access': ['-p cpu', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 120,
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        # NB: Vega's MaxMemPerNode is set to 256000, but this MUST be a MB/MiB units mistake
                        # Most likely, it is 256 GB, so 256*1E9/(1024*1024) MiB
                        'mem_per_node': 244140  # in MiB
                    },
                    'descr': 'CPU partition Standard, see https://en-doc.vega.izum.si/architecture/'
                },
                {
                    'name': 'gpu',
                    'scheduler': 'slurm',
                    'prepare_cmds': [
                        common_eessi_init(),
                        # Pass job environment variables like $PATH, etc., into job steps
                        'export SLURM_EXPORT_ENV=ALL',
                        # Needed when using srun launcher
                        # 'export SLURM_MPI_TYPE=pmix',  # WARNING: this broke the GROMACS on Vega
                        # Avoid https://github.com/EESSI/software-layer/issues/136
                        # Can be taken out once we don't care about old OpenMPI versions anymore (pre-4.1.1)
                        'export OMPI_MCA_pml=ucx',
                    ],
                    'launcher': 'mpirun',
                    # Use --export=None to avoid that login environment is passed down to submitted jobs
                    'access': ['-p gpu', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 60,
                    'devices': [
                        {
                            'type': DEVICE_TYPES[GPU],
                            'num_devices': 4,
                        }
                    ],
                    'resources': [
                        {
                            'name': '_rfm_gpu',
                            'options': ['--gpus-per-node={num_gpus_per_node}'],
                        },
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[GPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        GPU_VENDOR: GPU_VENDORS[NVIDIA],
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 476.837 * 1024  # in MiB (should be checked, its unclear from slurm.conf)
                    },
                    'descr': 'GPU partition, see https://en-doc.vega.izum.si/architecture/'
                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config(reframe_prefix)
        }
    ],
}



# ReFrame configuration file for VSC Tier-1 Hortense
# https://docs.vscentrum.be/en/latest/gent/tier1_hortense.html
#
# authors: Samuel Moors (VUB-HPC), Kenneth Hoste (HPC-UGent), Lara Peeters (HPC-UGent)

# Use generated topology file by ReFrame for CPU partitions
# `sched_access_in_submit` does not work with setting `'remote_detect': True,`

# Instructions on generating topology file
# ```
#    module swap cluster/{partition}
#    qsub -I -l nodes=1:ppn=all -l walltime=00:30:00
#
#    python3 -m venv "$TMPDIR"/reframe_venv
#    source "$TMPDIR"/reframe_venv/bin/activate
#    python3 -m pip install --upgrade pip
#    python3 -m pip install reframe-hpc=="4.6.2"
#
#    mkdir -p ~/.reframe/topology/hortense-{partition_name}
#    reframe --detect-host-topology \
#        ~/.reframe/topology/hortense-{partition_name}/processor.json
# ```
import os

from reframe.core.backends import register_launcher
from reframe.core.launchers import JobLauncher

from eessi.testsuite.common_config import (common_eessi_init,
                                           common_general_config,
                                           common_logging_config)
from eessi.testsuite.constants import *  # noqa: F403

# Note that we rely on the SBATCH_ACCOUNT environment variable to be specified
hortense_access = ['--export=NONE', '--get-user-env=60L']


@register_launcher('mympirun')
class MyMpirunLauncher(JobLauncher):
    def command(self, job):
        return ['mympirun', '--hybrid', str(job.num_tasks_per_node)]


eessi_cvmfs_repo = os.getenv('EESSI_CVMFS_REPO', None)
if eessi_cvmfs_repo is not None:
    prepare_eessi_init = "module --force purge"
    launcher = "mpirun"
    mpi_module = "env/vsc/dodrio/{}"
else:
    prepare_eessi_init = ""
    launcher = "mympirun"
    mpi_module = "vsc-mympirun"

site_configuration = {
    'systems': [
        {
            'name': 'hortense',
            'descr': 'Hortense',
            'hostnames': ['login.*.dodrio.os'],
            'modules_system': 'lmod',
            'partitions': [
                {
                    'name': 'cpu_rome',
                    'scheduler': 'slurm',
                    'prepare_cmds': [prepare_eessi_init, common_eessi_init()],
                    'access': hortense_access + ['--partition=cpu_rome'],
                    'sched_options': {
                        'sched_access_in_submit': True,
                    },
                    'environs': ['default'],
                    'descr': 'CPU nodes (AMD Rome, 256GiB RAM)',
                    'max_jobs': 20,
                    'launcher': launcher,
                    'modules': [mpi_module.format('cpu_rome')],
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 252160,  # in MiB
                    },
                },
                {
                    'name': 'cpu_rome_512',
                    'scheduler': 'slurm',
                    'prepare_cmds': [prepare_eessi_init, common_eessi_init()],
                    'access': hortense_access + ['--partition=cpu_rome_512'],
                    'sched_options': {
                        'sched_access_in_submit': True,
                    },
                    'environs': ['default'],
                    'descr': 'CPU nodes (AMD Rome, 512GiB RAM)',
                    'max_jobs': 20,
                    'launcher': launcher,
                    'modules': [mpi_module.format('cpu_rome_512')],
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 508160,  # in MiB
                    },
                },
                {
                    'name': 'cpu_milan',
                    'scheduler': 'slurm',
                    'prepare_cmds': [prepare_eessi_init, common_eessi_init()],
                    'access': hortense_access + ['--partition=cpu_milan'],
                    'sched_options': {
                        'sched_access_in_submit': True,
                    },
                    'environs': ['default'],
                    'descr': 'CPU nodes (AMD Milan, 256GiB RAM)',
                    'max_jobs': 20,
                    'launcher': launcher,
                    'modules': [mpi_module.format('cpu_milan')],
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 252160,  # in MiB
                    },
                },
                {
                    'name': 'gpu_rome_a100_40',
                    'scheduler': 'slurm',
                    'prepare_cmds': [prepare_eessi_init, common_eessi_init()],
                    'access': hortense_access + ['--partition=gpu_rome_a100_40'],
                    'sched_options': {
                        'sched_access_in_submit': True,
                    },
                    'environs': ['default'],
                    'descr': 'GPU nodes (A100 40GB)',
                    'max_jobs': 20,
                    'launcher': launcher,
                    'modules': [mpi_module.format('gpu_rome_a100_40')],
                    'features': [
                        FEATURES[GPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        GPU_VENDOR: GPU_VENDORS[NVIDIA],
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 254400,  # in MiB
                    },
                    'resources': [
                        {
                            'name': '_rfm_gpu',
                            'options': ['--gpus-per-node={num_gpus_per_node}'],
                        },
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'devices': [
                        {
                            'type': DEVICE_TYPES[GPU],
                            'num_devices': 4,
                        }
                    ],

                },
                {
                    'name': 'gpu_rome_a100_80',
                    'scheduler': 'slurm',
                    'prepare_cmds': [prepare_eessi_init, common_eessi_init()],
                    'access': hortense_access + ['--partition=gpu_rome_a100_80'],
                    'sched_options': {
                        'sched_access_in_submit': True,
                    },
                    'environs': ['default'],
                    'descr': 'GPU nodes (A100 80GB)',
                    'max_jobs': 20,
                    'launcher': launcher,
                    'modules': [mpi_module.format('gpu_rome_a100_80')],
                    'features': [
                        FEATURES[GPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        GPU_VENDOR: GPU_VENDORS[NVIDIA],
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 510720,  # in MiB
                    },
                    'resources': [
                        {
                            'name': '_rfm_gpu',
                            'options': ['--gpus-per-node={num_gpus_per_node}'],
                        },
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'devices': [
                        {
                            'type': DEVICE_TYPES[GPU],
                            'num_devices': 4,
                        }
                    ],

                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'gcc',
            'cxx': 'g++',
            'ftn': 'gfortran',
        },
        {
            'name': 'foss-2021a',
            'cc': 'mpicc',
            'cxx': 'mpicxx',
            'ftn': 'mpif90',
            'modules': ['foss/2021a']
        },
        {
            'name': 'intel-2021a',
            'modules': ['intel'],
            'cc': 'mpiicc',
            'cxx': 'mpiicpc',
            'ftn': 'mpiifort',
        },
        {
            'name': 'CUDA',
            'modules': ['CUDA'],
            'cc': 'nvcc',
            'cxx': 'nvcc',
        },
    ],
    'general': [
        {
            'remote_detect': True,
            'purge_environment': True,
            'resolve_module_conflicts': False,  # avoid loading the module before submitting the job
            **common_general_config()
        }
    ],
    'logging': common_logging_config(),
}



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914

import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import FEATURES, SCALES

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# AWS CITC site configuration
site_configuration = {
    'systems': [
        {
            'name': 'Magic_Castle',
            'descr': 'Magic Castle build and test environment on AWS',
            'modules_system': 'lmod',
            'hostnames': ['login*', '*-node'],
            'prefix': reframe_prefix,
            'partitions': [
                {
                    'name': 'x86_64-generic-16c-30gb',
                    'access': ['--partition=x86-64-generic-node', '--export=NONE'],
                    'descr': 'Generic (Haswell), 16 cores, 30 GB',
                },
                {
                    'name': 'x86_64-haswell-16c-30gb',
                    'access': ['--partition=x86-64-intel-haswell-node', '--export=NONE'],
                    'descr': 'Haswell, 16 cores, 30 GB',
                },
                {
                    'name': 'x86_64-skylake-16c-30gb',
                    'access': ['--partition=x86-64-intel-skylake-node', '--export=NONE'],
                    'descr': 'Skylake, 16 cores, 30 GB',
                },
                {
                    'name': 'x86_64-zen2-16c-30gb',
                    'access': ['--partition=x86-64-amd-zen2-node', '--export=NONE'],
                    'descr': 'Zen2, 16 cores, 30 GB',
                },
                {
                    'name': 'x86_64-zen3-16c-30gb',
                    'access': ['--partition=x86-64-amd-zen3-node', '--export=NONE'],
                    'descr': 'Zen3, 16 cores, 30 GiB',
                },
                {
                    'name': 'aarch64-generic-16c-32gb',
                    'access': ['--partition=aarch64-generic-node', '--export=NONE'],
                    'descr': 'Generic (Neoverse N1), 16 cores, 32 GB',
                },
                {
                    'name': 'aarch64-neoverse-V1-16c-32gb',
                    'access': ['--partition=aarch64-neoverse-v1-node', '--export=NONE'],
                    'descr': 'Neoverse V1, 16 cores, 32 GB',
                },
                {
                    'name': 'aarch64-neoverse-N1-16c-32gb',
                    'access': ['--partition=aarch64-neoverse-n1-node', '--export=NONE'],
                    'descr': 'Neoverse N1, 16 cores, 32 GiB',
                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config(reframe_prefix)
        }
    ],
}

# Add default things to each partition:
partition_defaults = {
    'scheduler': 'slurm',
    'launcher': 'mpirun',
    'environs': ['default'],
    'features': [
        FEATURES['CPU']
    ] + list(SCALES.keys()),
    'prepare_cmds': [
        common_eessi_init(),
        # Required when using srun as launcher with --export=NONE in partition access, in order to ensure job
        # steps inherit environment. It doesn't hurt to define this even if srun is not used
        'export SLURM_EXPORT_ENV=ALL'
    ],
    'resources': [
        {
            'name': 'memory',
            'options': ['--mem={size}'],
        }
    ],
    'extras': {
        # Node types have somewhat varying amounts of memory, but we'll make it easy on ourselves
        # All should _at least_ have this amount (30GB * 1E9 / (1024*1024) = 28610 MiB)
        'mem_per_node': 28610
    },
    'max_jobs': 1,
}
for system in site_configuration['systems']:
    for partition in system['partitions']:
        partition.update(partition_defaults)



"""
This module tests mpi4py's MPI_Reduce call
"""

import reframe as rfm
import reframe.utility.sanity as sn

# added only to make the linter happy
from reframe.core.builtins import variable, parameter, performance_function, sanity_function

# Import the EESSI_Mixin class so that we can inherit from it
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.constants import COMPUTE_UNIT, DEVICE_TYPES, CPU
from eessi.testsuite.utils import find_modules


# This python decorator indicates to ReFrame that this class defines a test
# Our class inherits from rfm.RunOnlyRegressionTest, since this test does not have a compilation stage
# https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RunOnlyRegressionTest
@rfm.simple_test
class EESSI_MPI4PY(rfm.RunOnlyRegressionTest, EESSI_Mixin):

    # The device type makes sure this test only gets executed on systems/partitions that can provide this device
    device_type = DEVICE_TYPES[CPU]

    # One task is launched per compute unit. In this case, one task per (physical) CPU core
    compute_unit = COMPUTE_UNIT[CPU]

    # ReFrame will generate a test for each module that matches the regex `mpi4py`
    # This means we implicitly assume that any module matching this name provides the required functionality
    # to run this test
    module_name = parameter(find_modules('mpi4py'))

    # Our script has two arguments, --n_iter and --n_warmup. By defining these as ReFrame variables, we can
    # enable the end-user to overwrite their value on the command line when invoking ReFrame.
    # Note that we don't typically expose ALL variables, especially if a script has many - we expose
    # only those that we think an end-user might want to overwrite
    # Number of iterations to run (more iterations takes longer, but results in more accurate timing)
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.variable
    n_iterations = variable(int, value=1000)

    # Similar for the number of warmup iterations
    n_warmup = variable(int, value=100)

    # Define which executable to run
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.executable
    executable = 'python3'

    # Define which options to pass to the executable
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.executable_opts
    executable_opts = ['mpi4py_reduce.py', '--n_iter', f'{n_iterations}', '--n_warmup', f'{n_warmup}']

    # Define a time limit for the scheduler running this test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.time_limit
    time_limit = '5m00s'

    # Define the benchmarks that are available in the test.
    # In this test (`EESSI_MPI4PY`) there is only one benchmark. If there are more than one,
    # define them using the `parameter()` function.
    bench_name = 'mpi4pi'

    # Specify the benchmark to be tested in CI (will be marked with a `CI` tag).
    bench_name_ci = 'mpi4pi'

    # Define the files and/or dirs inside sourcesdir (default=src) that should be symlinked into the stage dir
    readonly_files = ['mpi4py_reduce.py']

    # Define the class method that returns the required memory per node
    def required_mem_per_node(self):
        return self.num_tasks_per_node * 100 + 250

    # Now, we check if the pattern 'Sum of all ranks: X' with X the correct sum for the amount of ranks is found
    # in the standard output:
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.sanity_function
    @sanity_function
    def validate(self):
        # Sum of 0, ..., N-1 is (N * (N-1) / 2)
        sum_of_ranks = round(self.num_tasks * ((self.num_tasks - 1) / 2))
        # https://reframe-hpc.readthedocs.io/en/stable/deferrable_functions_reference.html#reframe.utility.sanity.assert_found
        return sn.assert_found(r'Sum of all ranks: %s' % sum_of_ranks, self.stdout)

    # Now, we define a pattern to extract a number that reflects the performance of this test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.performance_function
    @performance_function('s')
    def time(self):
        # https://reframe-hpc.readthedocs.io/en/stable/deferrable_functions_reference.html#reframe.utility.sanity.extractsingle
        return sn.extractsingle(r'^Time elapsed:\s+(?P<perf>\S+)', self.stdout, 'perf', float)



# ReFrame configuration file that can be used in GitHub Actions with EESSI

from eessi.testsuite.common_config import common_logging_config
from eessi.testsuite.constants import *


site_configuration = {
    'systems': [
        {
            'name': 'github_actions_eessi',
            'descr': 'GitHub Actions + EESSI',
            'hostnames': ['.*'],
            'modules_system': 'lmod',
            'partitions': [
                {
                    'name': 'default',
                    'scheduler': 'local',
                    'launcher': 'local',
                    'environs': ['default'],
                    'features': [FEATURES[CPU]] + list(SCALES.keys()),
                    'processor': {
                        'num_cpus': 2,
                        'num_cpus_per_core': 1,
                    },
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'max_jobs': 1,
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        # This is a fictional amount, GH actions probably has less, but only does --dry-run
                        'mem_per_node': 30 * 1024  # in MiB
                    },
                }
            ]
        }
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': ''
        }
    ],
    'general': [
        {
            'purge_environment': True,
            'resolve_module_conflicts': False,  # avoid loading the module before submitting the job
        }
    ],
    'logging': common_logging_config(),
}



import reframe as rfm
from reframe.core.builtins import parameter, run_after, performance_function, sanity_function
import reframe.utility.sanity as sn

from eessi.testsuite.constants import SCALES, COMPUTE_UNIT, DEVICE_TYPES, CPU
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.utils import find_modules


@rfm.simple_test
class EESSI_CP2K(rfm.RunOnlyRegressionTest, EESSI_Mixin):

    benchmark_info = parameter([
        # (bench_name, energy_ref, energy_tol)
        ('QS/H2O-32', -550.5055, 1e-4),
        ('QS/H2O-128', -2202.1791, 1e-4),
        ('QS/H2O-512', -8808.1439, 1e-4),
    ], fmt=lambda x: x[0], loggable=True)

    module_name = parameter(find_modules('CP2K'))
    scale = parameter(SCALES.keys())

    executable = 'cp2k.popt'
    time_limit = '2h'
    device_type = DEVICE_TYPES[CPU]
    compute_unit = COMPUTE_UNIT[CPU]
    bench_name_ci = 'QS/H2O-32'  # set CI on smallest benchmark
    readonly_files = ['QS']

    def required_mem_per_node(self):
        mems = {
            'QS/H2O-32': {'intercept': 0.5, 'slope': 0.15},
            'QS/H2O-128': {'intercept': 5, 'slope': 0.15},
            'QS/H2O-512': {'intercept': 34, 'slope': 0.20},
        }
        mem = mems[self.bench_name]
        return (self.num_tasks_per_node * mem['slope'] + mem['intercept']) * 1024

    @run_after('init')
    def set_bench_name(self):
        self.bench_name, self.energy_ref, self.energy_tol = self.benchmark_info
        self.descr = f'EESSI_CP2K {self.bench_name} benchmark'

    @run_after('setup')
    def prepare_test(self):
        self.executable_opts += ['-i', f'{self.bench_name}.inp']

    @sanity_function
    def assert_energy(self):
        energy = sn.extractsingle(
            r"^\s*ENERGY\| Total FORCE_EVAL.+?:\s*(?P<energy>.+)\n",
            self.stdout, 'energy', float, item=-1)
        energy_diff = sn.abs(energy - self.energy_ref)
        return sn.all([
            sn.assert_found(r'PROGRAM STOPPED IN', self.stdout),
            sn.assert_lt(energy_diff, self.energy_tol)
        ])

    @performance_function('s', perf_key='time')
    def time(self):
        return sn.extractsingle(r'^ CP2K(\s+[\d\.]+){4}\s+(?P<time>\S+)', self.stdout, 'time', float)

    @run_after('setup')
    def skip_tests(self):
        """Skip tests that are not suited for the requested resources"""

        # Skip QS/H2O-512 benchmark if not enough cores requested
        min_cores = 16
        self.skip_if(self.bench_name == 'QS/H2O-512' and self.num_tasks < min_cores,
                     f'Skipping benchmark {self.bench_name}: less than {min_cores} cores requested ({self.num_tasks})')

        # Skip QS/H2O-512 benchmark if too many nodes requested
        max_nodes = 8
        self.skip_if(self.bench_name == 'QS/H2O-512' and self.num_nodes > max_nodes,
                     f'Skipping benchmark {self.bench_name}: more than {max_nodes} nodes requested ({self.num_nodes})')



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914

import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import FEATURES, SCALES

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# AWS CITC site configuration
site_configuration = {
    'systems': [
        {
            'name': 'Magic_Castle_Azure',
            'descr': 'Magic Castle build and test environment on Azure',
            'modules_system': 'lmod',
            'hostnames': ['login.*', '.*-node'],
            'prefix': reframe_prefix,
            'partitions': [
                {
                    'name': 'x86_64-amd-zen4-node',
                    'access': ['--partition=x86-64-amd-zen4-node', '--export=NONE'],
                    'descr': 'Zen4, 16 cores, 30 GB',
                    'prepare_cmds': [
                        # Avoid
                        # https://www.eessi.io/docs/known_issues/eessi-2023.06/#eessi-production-repository-v202306
                        'export OMPI_MCA_btl=^uct,ofi'
                        'export OMPI_MCA_pml=ucx'
                        'export OMPI_MCA_mtl=^ofi'
                        # Use override to avoid fallback to zen3
                        'export EESSI_SOFTWARE_SUBDIR_OVERRIDE=x86_64/amd/zen4',
                        common_eessi_init(),
                        # Required when using srun as launcher with --export=NONE in partition access,
                        # in order to ensure job steps inherit environment. It doesn't hurt to define
                        # this even if srun is not used
                        'export SLURM_EXPORT_ENV=ALL'
                    ],
                    'extras': {
                        # For some reason, we cannot ask for the full amount configured as RealMemory in
                        # /etc/slurm/nodes.conf, so we ask slightly less
                        'mem_per_node': 767480
                    },
                },
                {
                    'name': 'aarch64-neoverse-N1-16c-62gb',
                    'access': ['--partition=aarch64-neoverse-n1-node', '--export=NONE'],
                    'descr': 'Neoverse N1, 16 cores, 62 GiB',
                    'prepare_cmds': [
                        common_eessi_init(),
                        # Required when using srun as launcher with --export=NONE in partition access,
                        # in order to ensure job steps inherit environment. It doesn't hurt to define
                        # this even if srun is not used
                        'export SLURM_EXPORT_ENV=ALL'
                    ],
                    'extras': {
                        # For some reason, we cannot ask for the full amount configured as RealMemory in
                        # /etc/slurm/nodes.conf, so we ask slightly less
                        'mem_per_node': 63480
                    },
                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config(reframe_prefix)
        }
    ],
}

# Add default things to each partition:
partition_defaults = {
    'scheduler': 'slurm',
    'launcher': 'mpirun',
    'environs': ['default'],
    'features': [
        FEATURES['CPU']
    ] + list(SCALES.keys()),
    'resources': [
        {
            'name': 'memory',
            'options': ['--mem={size}'],
        }
    ],
    'max_jobs': 1,
}
for system in site_configuration['systems']:
    for partition in system['partitions']:
        partition.update(partition_defaults)



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository
# 3. Temporarily change the 'access' field for the GPU partition to
# 'access':  ['-p gpu', '--export=None', '--gres=gpu:1'],

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job
# 3. Vega doesn't allow submission to the GPU partition without requesting at least one GPU (change #2)

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914

import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import *  # noqa: F403

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# This is an example configuration file
site_configuration = {
    'systems': [
        {
            'name': 'karolina',
            'descr': 'Karolina, a EuroHPC JU system',
            'modules_system': 'lmod',
            'hostnames': ['login*', 'acn*', 'cn*'],
            'prefix': reframe_prefix,
            'partitions': [
                {
                    'name': 'qcpu',
                    'scheduler': 'slurm',
                    'prepare_cmds': [
                        common_eessi_init(),
                        # Pass job environment variables like $PATH, etc., into job steps
                        'export SLURM_EXPORT_ENV=ALL',
                        # Needed when using srun launcher
                        # 'export SLURM_MPI_TYPE=pmix',  # WARNING: this broke the GROMACS on Vega
                        # Avoid https://github.com/EESSI/software-layer/issues/136
                        # Can be taken out once we don't care about old OpenMPI versions anymore (pre-4.1.1)
                        'export OMPI_MCA_pml=ucx',
                        # Work around "Failed to modify UD QP to INIT on mlx5_0: Operation not permitted" issue
                        # until we can resolve this through an LMOD hook in host_injections.
                        # (then these OMPI_MCA_btl & mtl can be removed again)
                        # See https://github.com/EESSI/software-layer/issues/456#issuecomment-2107755266
                        'export OMPI_MCA_mtl="^ofi"',
                        'export OMPI_MCA_btl="^ofi"',
                    ],
                    'launcher': 'mpirun',
                    # Use --export=None to avoid that login environment is passed down to submitted jobs
                    # Note that we rely on the SBATCH_ACCOUNT environment variable to be specified
                    'access': ['-p qcpu', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 120,
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 235520  # in MiB
                    },
                    'descr': 'CPU Universal Compute Nodes, see https://docs.it4i.cz/karolina/hardware-overview/'
                },
                # We don't have GPU budget on Karolina at this time
                # {
                #     'name': 'qgpu',
                #     'scheduler': 'slurm',
                #     'prepare_cmds': [
                #         common_eessi_init(),
                #         # Pass job environment variables like $PATH, etc., into job steps
                #         'export SLURM_EXPORT_ENV=ALL',
                #         # Needed when using srun launcher
                #         # 'export SLURM_MPI_TYPE=pmix',  # WARNING: this broke the GROMACS on Vega
                #         # Avoid https://github.com/EESSI/software-layer/issues/136
                #         # Can be taken out once we don't care about old OpenMPI versions anymore (pre-4.1.1)
                #         'export OMPI_MCA_pml=ucx',
                #     ],
                #     'launcher': 'mpirun',
                #     # Use --export=None to avoid that login environment is passed down to submitted jobs
                #     # Note that we rely on the SBATCH_ACCOUNT environment variable to be specified
                #     'access':  ['-p gpu', '--export=None'],
                #     'environs': ['default'],
                #     'max_jobs': 60,
                #     'devices': [
                #         {
                #             'type': DEVICE_TYPES[GPU],
                #             'num_devices': 8,
                #         }
                #     ],
                #     'resources': [
                #         {
                #             'name': '_rfm_gpu',
                #             'options': ['--gpus-per-node={num_gpus_per_node}'],
                #         }
                #     ],
                #     'features': [
                #         FEATURES[GPU],
                #     ] + list(SCALES.keys()),
                #     'descr': 'GPU partition with accelerated nodes, https://docs.it4i.cz/karolina/hardware-overview/'
                # },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config(reframe_prefix)
        }
    ],
}



"""
This module tests mpi4py's MPI_Reduce call
"""

import reframe as rfm
import reframe.utility.sanity as sn

# added only to make the linter happy
from reframe.core.builtins import variable, parameter, run_after, performance_function, sanity_function


# This python decorator indicates to ReFrame that this class defines a test
# Our class inherits from rfm.RunOnlyRegressionTest, since this test does not have a compilation stage
# https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RunOnlyRegressionTest
@rfm.simple_test
class EESSI_MPI4PY(rfm.RunOnlyRegressionTest):
    # Programming environments are only relevant for tests that compile something
    # Since we are testing existing modules, we typically don't compile anything and simply define
    # 'default' as the valid programming environment
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.valid_prog_environs
    valid_prog_environs = ['default']

    # Typically, we list here the name of our cluster as it is specified in our ReFrame configuration file
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.valid_systems
    valid_systems = ['snellius']

    # ReFrame will generate a test for each module
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.parameter
    module_name = parameter(['mpi4py/3.1.4-gompi-2023a', 'mpi4py/3.1.5-gompi-2023b'])

    # ReFrame will generate a test for each scale
    scale = parameter([2, 128, 256])

    # Our script has two arguments, --n_iter and --n_warmup. By defining these as ReFrame variables, we can
    # enable the end-user to overwrite their value on the command line when invoking ReFrame.
    # Note that we don't typically expose ALL variables, especially if a script has many - we expose
    # only those that we think an end-user might want to overwrite
    # Number of iterations to run (more iterations takes longer, but results in more accurate timing)
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.variable
    n_iterations = variable(int, value=1000)

    # Similar for the number of warmup iterations
    n_warmup = variable(int, value=100)

    # Define which executable to run
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.executable
    executable = 'python3'

    # Define which options to pass to the executable
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.executable_opts
    executable_opts = ['mpi4py_reduce.py', '--n_iter', f'{n_iterations}', '--n_warmup', f'{n_warmup}']

    # Define a time limit for the scheduler running this test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.time_limit
    time_limit = '5m00s'

    # Using this decorator, we tell ReFrame to run this AFTER the init step of the test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.run_after
    # See https://reframe-hpc.readthedocs.io/en/stable/pipeline.html for all steps in the pipeline
    # that reframe uses to execute tests. Note that after the init step, ReFrame has generated test instances for each
    # of the combinations of parameters above. Thus, now, there are 6 instances (2 module names * 3 scales). Here,
    # we set the modules to load equal to one of the module names
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.modules
    @run_after('init')
    def set_modules(self):
        self.modules = [self.module_name]

    # Similar for the scale, we now set the number of tasks equal to the scale for this instance
    @run_after('init')
    def define_task_count(self):
        # Set the number of tasks, self.scale is now a single number out of the parameter list
        # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.num_tasks
        self.num_tasks = self.scale
        # Set the number of tasks per node to either be equal to the number of tasks, but at most 128,
        # since we have 128-core nodes
        # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.num_tasks_per_node
        self.num_tasks_per_node = min(self.num_tasks, 128)

    # Now, we check if the pattern 'Sum of all ranks: X' with X the correct sum for the amount of ranks is found
    # in the standard output:
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.sanity_function
    @sanity_function
    def validate(self):
        # Sum of 0, ..., N-1 is (N * (N-1) / 2)
        sum_of_ranks = round(self.num_tasks * ((self.num_tasks - 1) / 2))
        # https://reframe-hpc.readthedocs.io/en/stable/deferrable_functions_reference.html#reframe.utility.sanity.assert_found
        return sn.assert_found(r'Sum of all ranks: %s' % sum_of_ranks, self.stdout)

    # Now, we define a pattern to extract a number that reflects the performance of this test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.performance_function
    @performance_function('s')
    def time(self):
        # https://reframe-hpc.readthedocs.io/en/stable/deferrable_functions_reference.html#reframe.utility.sanity.extractsingle
        return sn.extractsingle(r'^Time elapsed:\s+(?P<perf>\S+)', self.stdout, 'perf', float)



import os

import reframe.core.logging as rflog

perflog_format = '|'.join([
    '%(check_job_completion_time)s',
    '%(osuser)s',
    '%(version)s',
    '%(check_unique_name)s',
    '%(check_info)s',
    '%(check_system)s',
    '%(check_partition)s',
    '%(check_environ)s',
    '%(check_exclusive_access)s',
    '%(check_num_tasks)s',
    '%(check_num_cpus_per_task)s',
    '%(check_num_tasks_per_node)s',
    '%(check_num_gpus_per_node)s',
    '%(check_use_multithreading)s',
    '%(check_modules)s',
    '%(check_jobid)s',
    '%(check_perfvalues)s',
])

format_perfvars = '|'.join([
    '%(check_perf_var)s',
    '%(check_perf_value)s',
    '%(check_perf_lower_thres)s',
    '%(check_perf_upper_thres)s',
    '%(check_perf_unit)s',
    ''  # final delimiter required
])


def common_logging_config(prefix=None):
    """
    return default logging configuration as a list: stdout, file log, perflog
    :param prefix: file log prefix
    """
    prefix = os.getenv('RFM_PREFIX', prefix if prefix else '.')
    logdir = os.path.join(prefix, 'logs')
    os.makedirs(logdir, exist_ok=True)

    return [{
        'level': 'debug',
        'handlers': [
            {
                'type': 'stream',
                'name': 'stdout',
                'level': 'info',
                'format': '%(message)s',
            },
            {
                'type': 'file',
                'name': os.path.join(logdir, 'reframe.log'),
                'level': 'debug',
                'format': '[%(asctime)s] %(levelname)s: %(check_info)s: %(message)s',
                'append': True,
                'timestamp': "%Y%m%d_%H%M%S",  # add a timestamp to the filename (reframe_<timestamp>.log)
            },
        ],
        'handlers_perflog': [
            {
                'type': 'filelog',
                'prefix': '%(check_system)s/%(check_partition)s',
                'level': 'info',
                'format': perflog_format,
                'format_perfvars': format_perfvars,
                'append': True,  # avoid overwriting
            },
        ],
    }]


def common_general_config(prefix=None):
    """
    return common configuration for the 'general' section of the ReFrame configuration file
    :param: prefix: prefix for the report_file
    """
    prefix = os.getenv('RFM_PREFIX', prefix if prefix else '.')
    reportdir = os.path.join(prefix, 'report_files')
    os.makedirs(reportdir, exist_ok=True)

    return {
        'check_search_recursive': True,
        'report_file': os.path.join(reportdir, 'run-report-{sessionid}.json')
    }


def common_eessi_init(eessi_version=None):
    """
    Returns the full path that should be sourced to initialize the EESSI environment for a given version of EESSI.
    If no eessi_version is passed, the EESSI_VERSION environment variable is read.
    If that is also not defined, default behaviour is to use `latest`.
    :param eessi_version: version of EESSI that should be sourced (e.g. '2023.06' or 'latest') [optional]
    """
    # Check which EESSI_CVMFS_REPO we are running under
    eessi_cvmfs_repo = os.getenv('EESSI_CVMFS_REPO', None)

    if eessi_cvmfs_repo is None:
        rflog.getlogger().warning(' '.join([
            "Environment variable 'EESSI_CVMFS_REPO' is not defined.",
            "If you do not intend to use the EESSI software stack, this is perfectly fine.",
            "To use EESSI, initialize the EESSI environment before running the test suite.",
        ]))
        return ''

    eessi_init = []
    pilot_repo = '/cvmfs/pilot.eessi-hpc.org'

    if eessi_cvmfs_repo == pilot_repo:
        eessi_init.append('export EESSI_FORCE_PILOT=1')
        if eessi_version is None:
            # Try also EESSI_VERSION for backwards compatibility with previous common_eessi_init implementation
            eessi_version = os.getenv('EESSI_PILOT_VERSION', os.getenv('EESSI_VERSION', 'latest'))
    else:
        # software.eessi.io, or another where we assume the same variable names to be used
        if eessi_version is None:
            eessi_version = os.getenv('EESSI_VERSION', None)
        # Without EESSI_VERSION, we don't know what to do. There is no default/latest version
        # So, report error
        if eessi_version is None:
            err_msg = "Environment variable 'EESSI_VERSION' was not found."
            err_msg += " Did you initialize the EESSI environment before running the test suite?"
            raise ValueError(err_msg)

    if eessi_cvmfs_repo == pilot_repo and eessi_version == 'latest':
        version_string = eessi_version
    else:
        version_string = f'versions/{eessi_version}'

    eessi_init.append(f'source {eessi_cvmfs_repo}/{version_string}/init/bash')
    return ' && '.join(eessi_init)



"""
Utility functions for ReFrame tests
"""

import re
import sys
from typing import Iterator

import reframe as rfm
import reframe.core.runtime as rt
from reframe.frontend.printer import PrettyPrinter

from eessi.testsuite.constants import *

printer = PrettyPrinter()


def log(msg, logger=printer.debug):
    funcname = sys._getframe().f_back.f_code.co_name
    logger(f'[{funcname}]: {msg}')


def _get_gpu_list(test: rfm.RegressionTest) -> list:
    return [dev.num_devices for dev in test.current_partition.devices if dev.device_type == DEVICE_TYPES[GPU]]


def get_max_avail_gpus_per_node(test: rfm.RegressionTest) -> int:
    """
    Returns the maximum available number of GPUs per node for the current partition,
    taken from 'num_devices' of device GPU_DEV_NAME in the 'devices' attribute of the current partition
    """
    gpu_list = _get_gpu_list(test)
    # If multiple devices have type DEVICE_TYPES[GPU] in the current partition,
    # we don't know for which to return the device count...
    if len(gpu_list) != 1:
        partname = test.current_partition.name
        raise ValueError(
            f"{len(gpu_list)} devices defined in config file with name {DEVICE_TYPES[GPU]} for partition {partname}. "
            "Cannot determine maximum number of GPUs available for the test. "
            f"Please check the definition of partition {partname} in your ReFrame config file."
        )

    return gpu_list[0]


def is_gpu_present(test: rfm.RegressionTest) -> bool:
    '''Checks if GPUs are present in the current partition'''
    return len(_get_gpu_list(test)) >= 1


def is_cuda_required_module(module_name: str) -> bool:
    '''Checks if CUDA seems to be required by given module'''
    requires_cuda = False
    if re.search("(?i)cuda", module_name):
        requires_cuda = True
    return requires_cuda


def find_modules(regex: str, name_only=True) -> Iterator[str]:
    """
    Return all modules matching the regular expression regex. Note that since we use re.search,
    a module matches if the regex matches the module name at any place. I.e. the match does
    not have to be at the start of the smodule name

    Arguments:
    - regex: a regular expression
    - name_only: regular expressions will only be matched on the module name, not the version (default: True).

    Note: the name_only feature assumes anything after the last forward '/' is the version,
    and strips that before doing a match.

    Example

    Suppose we have the following modules on a system:

    gompic/2022a
    gompi/2022a
    CGAL/4.14.3-gompi-2022a

    The following calls would return the following respective modules

    find_modules('gompi') => [gompic/2022a, gompi/2022a]
    find_modules('gompi$') => [gompi/2022a]
    find_modules('gompi', name_only = False) => [gompic/2022a, gompi/2022a, CGAL/4.14.3-gompi-2022a]
    find_modules('^gompi', name_only = False) => [gompic/2022a, gompi/2022a]
    find_modules('^gompi/', name_only = False) => [gompi/2022a]
    find_modules('-gompi-2022a', name_only = False) => [CGAL/4.14.3-gompi-2022a]

    """

    if not isinstance(regex, str):
        raise TypeError("'substr' argument must be a string")

    ms = rt.runtime().modules_system
    # Returns e.g. ['Bison/', 'Bison/3.7.6-GCCcore-10.3.0', 'BLIS/', 'BLIS/0.8.1-GCC-10.3.0']
    modules = ms.available_modules('')
    seen = set()
    dupes = []
    for mod in modules:
        # Exclude anything without version, i.e. ending with / (e.g. Bison/)
        if re.search('.*/$', mod):
            continue
        # The thing we yield should always be the original module name (orig_mod), including version
        orig_mod = mod
        if name_only:
            # Remove trailing slashes from the regex (in case the callee forgot)
            regex = regex.rstrip('/')
            # Remove part after the last forward slash, as we assume this is the version
            mod = re.sub('/[^/]*$', '', mod)
        # Match the actual regular expression
        log(f"Matching module {mod} with regex {regex}")
        if re.search(regex, mod):
            log("Match!")
            if orig_mod in seen:
                dupes.append(orig_mod)
            else:
                seen.add(orig_mod)
            yield orig_mod

    if dupes:
        err_msg = "EESSI test-suite cannot handle duplicate modules. "
        err_msg += "Please make sure that only one is available on your system. "
        err_msg += f"The following modules have a duplicate on your system: {dupes}"
        raise ValueError(err_msg)


def check_proc_attribute_defined(test: rfm.RegressionTest, attribute) -> bool:
    """
    Checks if a processor feature is defined (i.e. if test.current_partition.processor.<somefeature> is defined)
    If not, throws an informative error message.

    Arguments:
    - test: the reframe regression test instance for which should be checked if the processor feature is defined
    - attribute: attribute of the processor object, as defined by systems.partitions.processor

    Return:
    - True (bool) if the attribute is defined
    - Function does not return (but raises an error) if the attribute is undefined

    Current known attributes in ReFrame are arch, num_cpus, num_cpus_per_core and topology,
    but this may change in the future.

    If ReFrame's autodetect feature is used, all of these should be properly defined, so that's what we advice.
    """

    if test.current_partition:
        if getattr(test.current_partition.processor, attribute):
            return True
        else:
            msg = (
                f"Processor information ({attribute}) missing. "
                "Check that processor information is either autodetected "
                "(see https://reframe-hpc.readthedocs.io/en/stable/configure.html#proc-autodetection), "
                "or manually set in the ReFrame configuration file "
                "(see https://reframe-hpc.readthedocs.io/en/stable/config_reference.html#processor-info)."
            )
    else:
        msg = (
            "This test's current_partition is not set yet. "
            "The function utils.check_proc_attribute_defined should only be called after the setup() phase of ReFrame."
            "This is a programming error, please report this issue."
        )
    raise AttributeError(msg)


def check_extras_key_defined(test: rfm.RegressionTest, extra_key) -> bool:
    """
    Checks if a specific key is defined in the 'extras' dictionary for the current partition
    (i.e. if test.current_partition.extras[extra_key] is defined)
    If not, throws an informative error message.
    Note that partition extras are defined by free text keys, so any string is (potentially) valid.

    Arguments:
    - test: the reframe regression test instance for which should be checked if the key is defined in 'extras'
    - extra_key: key for which to check in the 'extras' dictionary

    Return:
    - True (bool) if the key is defined
    - Function does not return (but raises an error) if the attribute is undefined
    """

    if test.current_partition:
        if extra_key in test.current_partition.extras:
            return True
        else:
            msg = (
                f"Key '{extra_key}' missing in the 'extras' dictionary for partition '{test.current_partition.name}'."
                "Please define this key for the relevant partition in the ReFrame configuration file (see "
                "https://reframe-hpc.readthedocs.io/en/stable/config_reference.html#config.systems.partitions.extras)."
            )
    else:
        msg = (
            "This test's current_partition is not set yet. "
            "The function utils.check_extras_key_defined should only be called after the setup() phase of ReFrame."
            "This is a programming error, please report this issue."
        )
    raise AttributeError(msg)



"""
This module tests TensorFlow in available modules containing substring 'TensorFlow'.
The test itself is based on an official multi-worker with Keras tutoral at
https://www.tensorflow.org/tutorials/distribute/multi_worker_with_keras
"""

import reframe as rfm
from reframe.core.builtins import deferrable, parameter, run_after, sanity_function, performance_function
import reframe.utility.sanity as sn

from eessi.testsuite import utils
from eessi.testsuite.constants import COMPUTE_UNIT, CPU, CPU_SOCKET, DEVICE_TYPES, GPU
from eessi.testsuite.eessi_mixin import EESSI_Mixin


@rfm.simple_test
class EESSI_TensorFlow(rfm.RunOnlyRegressionTest, EESSI_Mixin):

    # Parameterize over all modules that start with TensorFlow
    module_name = parameter(utils.find_modules('TensorFlow'))

    # Make CPU and GPU versions of this test
    device_type = parameter([DEVICE_TYPES[CPU], DEVICE_TYPES[GPU]])

    executable = 'python tf_test.py'
    readonly_files = ['mnist_setup.py', 'tf_test.py']

    time_limit = '30m'

    # This test should be run as part of EESSI CI
    bench_name = bench_name_ci = 'bench_ci'

    readonly_files = ['mnist_setup.py', 'tf_test.py']

    def required_mem_per_node(self):
        return self.num_tasks_per_node * 2048

    @deferrable
    def assert_tf_config_ranks(self):
        '''Assert that each rank sets a TF_CONFIG'''
        n_ranks = sn.count(sn.extractall(
            '^Rank [0-9]+: Set TF_CONFIG for rank (?P<rank>[0-9]+)', self.stdout, tag='rank'))
        return sn.assert_eq(n_ranks, self.num_tasks)

    @deferrable
    def assert_completion(self):
        '''Assert that the test ran until completion'''
        n_fit_completed = sn.count(sn.extractall('^Rank [0-9]+: Keras fit completed', self.stdout))

        return sn.assert_eq(n_fit_completed, self.num_tasks)

    @deferrable
    def assert_convergence(self):
        '''Assert that the network learned _something_ during training'''
        accuracy = sn.extractsingle('^Final accuracy: (?P<accuracy>\S+)', self.stdout, 'accuracy', float)  # noqa: W605
        # mnist is a 10-class classification problem, so if accuracy >> 0.2 the network 'learned' something
        return sn.assert_gt(accuracy, 0.2)

    @sanity_function
    def assert_sanity(self):
        '''Check all sanity criteria'''
        return sn.all([
            self.assert_tf_config_ranks(),
            self.assert_completion(),
            self.assert_convergence(),
        ])

    @performance_function('img/s')
    def perf(self):
        return sn.extractsingle(r'^Performance:\s+(?P<perf>\S+)', self.stdout, 'perf', float)

    @run_after('init')
    def set_executable_opts(self):
        """Set executable opts based on device_type parameter"""
        self.executable_opts += ['--device', self.device_type]
        utils.log(f'executable_opts set to {self.executable_opts}')

    @run_after('init')
    def set_test_descr(self):
        self.descr = f'TensorFlow benchmark on {self.device_type}'

    @run_after('init')
    def set_compute_unit(self):
        """
        Set the compute unit to which tasks will be assigned:
        one task per CPU socket for CPU runs, and one task per GPU for GPU runs.
        """
        device_to_compute_unit = {
            DEVICE_TYPES[CPU]: COMPUTE_UNIT[CPU_SOCKET],
            DEVICE_TYPES[GPU]: COMPUTE_UNIT[GPU],
        }
        self.compute_unit = device_to_compute_unit.get(self.device_type)

    @run_after('setup')
    def set_thread_count_args(self):
        """Set executable opts defining the thread count"""
        self.executable_opts += ['--intra-op-parallelism', '%s' % self.num_cpus_per_task]
        self.executable_opts += ['--inter-op-parallelism', '1']
        utils.log(f'executable_opts set to {self.executable_opts}')



"""
This module tests the binary 'pw.x' in available modules containing substring 'QuantumESPRESSO'.
Test input files are defined in the ReFrame test library,
see https://github.com/reframe-hpc/reframe/blob/develop/hpctestlib/sciapps/qespresso/benchmarks.py

ReFrame terminology:

"pipeline stages":
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#pipeline-hooks

"test parameter": a list of values, which will generate different test variants.
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.parameter

"test variant": a version of a test with a specific value for each test parameter
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#test-variants

"concrete test cases": all test combinations that will actually run:
- test variants
- valid system:partition+programming environment combinations
https://reframe-hpc.readthedocs.io/en/stable/tutorial_deps.html#listing-dependencies

Tests can be filtered by name, tag, programming environment, system, partition, or maintainer,
see https://reframe-hpc.readthedocs.io/en/stable/manpage.html#test-filtering

Hooks acting on all possible test combinations (before filtering) are called after the 'init' stage.
Hooks acting on concrete test cases (after filtering) are called after the 'setup' stage.

See also https://reframe-hpc.readthedocs.io/en/stable/pipeline.html
"""

import reframe as rfm
from hpctestlib.sciapps.qespresso.benchmarks import QEspressoPWCheck
from reframe.core.builtins import parameter, run_after

from eessi.testsuite.constants import COMPUTE_UNIT, CPU, DEVICE_TYPES, GPU
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.utils import find_modules


@rfm.simple_test
class EESSI_QuantumESPRESSO_PW(QEspressoPWCheck, EESSI_Mixin):
    time_limit = '30m'
    module_name = parameter(find_modules('QuantumESPRESSO'))
    # For now, QE is built for CPU targets only
    device_type = parameter([DEVICE_TYPES[CPU]])
    readonly_files = ['']

    def required_mem_per_node(self):
        return (self.num_tasks_per_node * 0.9 + 4) * 1024

    @run_after('init')
    def set_ci(self):
        """Set tag CI on smallest benchmark, so it can be selected on the cmd line via --tag CI"""
        min_ecut = min(QEspressoPWCheck.ecut.values)
        min_nbnd = min(QEspressoPWCheck.nbnd.values)
        if self.ecut == min_ecut and self.nbnd == min_nbnd:
            self.bench_name = self.bench_name_ci = 'bench_ci'

    @run_after('init')
    def set_increased_walltime(self):
        """Increase the amount of time for the largest benchmark, so it can complete successfully."""
        max_ecut = max(QEspressoPWCheck.ecut.values)
        max_nbnd = max(QEspressoPWCheck.nbnd.values)
        if self.ecut == max_ecut and self.nbnd == max_nbnd:
            self.time_limit = '60m'

    @run_after('init')
    def set_compute_unit(self):
        """
        Set the compute unit to which tasks will be assigned:
        one task per CPU core for CPU runs, and one task per GPU for GPU runs.
        """
        device_to_compute_unit = {
            DEVICE_TYPES[CPU]: COMPUTE_UNIT[CPU],
            DEVICE_TYPES[GPU]: COMPUTE_UNIT[GPU],
        }
        self.compute_unit = device_to_compute_unit.get(self.device_type)



"""
This module tests the binary 'gmx_mpi' in available modules containing substring 'GROMACS'.
Test input files are taken from https://www.hecbiosim.ac.uk/access-hpc/benchmarks,
as defined in the ReFrame test library,
see https://github.com/reframe-hpc/reframe/blob/develop/hpctestlib/sciapps/gromacs/benchmarks.py

ReFrame terminology:

"pipeline stages":
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#pipeline-hooks

"test parameter": a list of values, which will generate different test variants.
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.parameter

"test variant": a version of a test with a specific value for each test parameter
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#test-variants

"concrete test cases": all test combinations that will actually run:
- test variants
- valid system:partition+programming environment combinations
https://reframe-hpc.readthedocs.io/en/stable/tutorial_deps.html#listing-dependencies

Tests can be filtered by name, tag, programming environment, system, partition, or maintainer,
see https://reframe-hpc.readthedocs.io/en/stable/manpage.html#test-filtering

Hooks acting on all possible test combinations (before filtering) are called after the 'init' stage.
Hooks acting on concrete test cases (after filtering) are called after the 'setup' stage.

See also https://reframe-hpc.readthedocs.io/en/stable/pipeline.html
"""

import reframe as rfm
from reframe.core.builtins import parameter, run_after  # added only to make the linter happy

from hpctestlib.sciapps.gromacs.benchmarks import gromacs_check

from eessi.testsuite.constants import COMPUTE_UNIT, DEVICE_TYPES, SCALES
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.utils import find_modules, log


class EESSI_GROMACS_base(gromacs_check):
    @run_after('init')
    def set_device_type(self):
        self.device_type = self.nb_impl


@rfm.simple_test
class EESSI_GROMACS(EESSI_GROMACS_base, EESSI_Mixin):
    scale = parameter(SCALES.keys())
    time_limit = '30m'
    module_name = parameter(find_modules('GROMACS'))
    bench_name_ci = 'HECBioSim/Crambin'
    # input files are downloaded
    readonly_files = ['']
    # executable_opts in addition to those set by the hpctestlib
    executable_opts = ['-dlb', 'yes', '-npme', '-1']

    def required_mem_per_node(self):
        return self.num_tasks_per_node * 1024

    def __init__(self):
        # self.device_type must be set before the @run_after('init') hooks of the EESSI_Mixin class
        self.device_type = self.nb_impl

    @run_after('init')
    def set_compute_unit(self):
        """
        Set the compute unit to which tasks will be assigned:
        one task per CPU core for CPU runs, and one task per GPU for GPU runs.
        """
        if self.device_type == DEVICE_TYPES['CPU']:
            self.compute_unit = COMPUTE_UNIT['CPU']
        elif self.device_type == DEVICE_TYPES['GPU']:
            self.compute_unit = COMPUTE_UNIT['GPU']
        else:
            msg = f"No mapping of device type {self.device_type} to a COMPUTE_UNIT was specified in this test"
            raise NotImplementedError(msg)

    @run_after('setup')
    def set_omp_num_threads(self):
        """
        Set number of OpenMP threads.
        Set both OMP_NUM_THREADS and -ntomp explicitly to avoid conflicting values.
        Set default number of OpenMP threads equal to number of CPUs per task.
        Also support setting OpenMP threads on the cmd line via custom executable option '-ntomp'.
        """
        omp_num_threads = self.num_cpus_per_task
        self.executable_opts += ['-ntomp', str(omp_num_threads)]
        log(f'executable_opts set to {self.executable_opts}')

        self.env_vars['OMP_NUM_THREADS'] = omp_num_threads
        log(f'env_vars set to {self.env_vars}')



"""
This module tests Espresso in available modules containing substring 'ESPResSo' which is different from Quantum
Espresso. Tests included:
- P3M benchmark - Ionic crystals
    - Weak scaling
    - Strong scaling Weak and strong scaling are options that are needed to be provided to the script and the system is
      either scaled based on number of cores or kept constant.
"""

import reframe as rfm
import reframe.utility.sanity as sn

from reframe.core.builtins import deferrable, parameter, performance_function, run_after, sanity_function
from reframe.utility import reframe

from eessi.testsuite.constants import CPU, DEVICE_TYPES, SCALES, COMPUTE_UNIT
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.utils import find_modules, log


def filter_scales():
    """
    Filtering function for filtering scales for P3M test and the LJ test.
    This is currently required because the 16 node test takes way too long and always fails due to time limit.
    Once a solution to mesh tuning algorithm is found, where we can specify the mesh sizes for a particular scale,
    this function can be removed.
    """
    return [
        k for (k, v) in SCALES.items()
        if v['num_nodes'] != 16
    ]


class EESSI_ESPRESSO_base(rfm.RunOnlyRegressionTest):
    module_name = parameter(find_modules('^ESPResSo$'))
    device_type = DEVICE_TYPES[CPU]
    compute_unit = COMPUTE_UNIT[CPU]
    time_limit = '300m'

    @run_after('init')
    def set_ci_tag(self):
        """ Setting tests under CI tag. """
        # this test runs longer at larger scales due to mesh tuning
        # thus, we only set CI tag on scales < 2 nodes to limit execution time
        # TODO: revisit this for more recent versions of ESPResSo
        # see also: https://github.com/EESSI/test-suite/issues/154
        if SCALES[self.scale]['num_nodes'] < 2:
            self.bench_name_ci = self.bench_name

    @sanity_function
    def assert_sanity(self):
        '''Check all sanity criteria'''
        return sn.all([
            self.assert_completion(),
            self.assert_convergence(),
        ])

    @performance_function('s/step')
    def perf(self):
        return sn.extractsingle(r'^Performance:\s+(?P<perf>\S+)', self.stdout, 'perf', float)


@rfm.simple_test
class EESSI_ESPRESSO_P3M_IONIC_CRYSTALS(EESSI_ESPRESSO_base, EESSI_Mixin):
    scale = parameter(filter_scales())

    executable = 'python3 madelung.py'
    sourcesdir = 'src/p3m'
    readonly_files = ['madelung.py']
    bench_name = 'ionic_crystals_p3m'

    default_weak_scaling_system_size = 6

    def required_mem_per_node(self):
        return (self.num_tasks_per_node * 0.9) * 1024

    @run_after('init')
    def set_executable_opts(self):
        """Set executable opts based on device_type parameter"""
        # Weak scaling (Gustafson's law: constant work per core): size scales with number of cores
        self.executable_opts += ['--size', str(self.default_weak_scaling_system_size), '--weak-scaling']
        log(f'executable_opts set to {self.executable_opts}')

    @deferrable
    def assert_completion(self):
        '''Check completion'''
        cao = sn.extractsingle(r'^resulting parameters:.*cao: (?P<cao>\S+),', self.stdout, 'cao', int)
        return (sn.assert_found(r'^Algorithm executed.', self.stdout) and cao)

    @deferrable
    def assert_convergence(self):
        '''Check convergence'''
        check_string = sn.assert_found(r'Final convergence met with tolerances:', self.stdout)
        energy = sn.extractsingle(r'^\s+energy:\s+(?P<energy>\S+)', self.stdout, 'energy', float)
        return (check_string and (energy != 0.0))


@rfm.simple_test
class EESSI_ESPRESSO_LJ_PARTICLES(EESSI_ESPRESSO_base, EESSI_Mixin):
    scale = parameter(filter_scales())

    executable = 'python3 lj.py'
    sourcesdir = 'src/lj'
    readonly_files = ['lj.py']
    bench_name = 'particles_lj'

    def required_mem_per_node(self):
        "LJ requires 200 MB per core"
        return (self.num_tasks_per_node * 0.3) * 1024

    @deferrable
    def assert_completion(self):
        '''Check completion'''
        return (sn.assert_found(r'^Algorithm executed.', self.stdout))

    @deferrable
    def assert_convergence(self):
        '''Check convergence'''
        check_string = sn.assert_found(r'Final convergence met with relative tolerances:', self.stdout)
        energy = sn.extractsingle(r'^\s+sim_energy:\s+(?P<energy>\S+)', self.stdout, 'energy', float)
        return (check_string and (energy != 0.0))



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository
# If your system has a GPU partition, it might force jobs to request at least one GPU. If that is the
# case, you also need to temporarily change 'access' field for the GPU partition to include the request
# for one GPU, e.g. 'access':  ['-p gpu', '--export=None', '--gres=gpu:1'],

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914


"""
Example configuration file
"""
import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import FEATURES, CPU, GPU, SCALES, DEVICE_TYPES, GPU_VENDOR, GPU_VENDORS, NVIDIA


site_configuration = {
    'systems': [
        {
            'name': 'example',
            'descr': 'Example cluster',
            'modules_system': 'lmod',
            'hostnames': ['.*'],
            # Note that the stagedir should be a shared directory available on all nodes running ReFrame tests
            'stagedir': f'/some/shared/dir/{os.environ.get("USER")}/reframe_output/staging',
            'partitions': [
                {
                    'name': 'cpu_partition',
                    'descr': 'CPU partition',
                    'scheduler': 'slurm',
                    'launcher': 'mpirun',
                    'access': ['-p cpu', '--export=None'],
                    'prepare_cmds': [common_eessi_init()],
                    'environs': ['default'],
                    'max_jobs': 4,
                    # We recommend to rely on ReFrame's CPU autodetection,
                    # and only define the 'processor' field if autodetection fails
                    # 'processor': {
                        # 'num_cpus': 128,
                        # 'num_sockets': 2,
                        # 'num_cpus_per_socket': 64,
                        # 'num_cpus_per_core': 1,
                    # },
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'extras': {
                        # If you have slurm, check with scontrol show node <nodename> for the amount of RealMemory
                        # on nodes in this partition
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 229376  # in MiB
                    },
                    # list(SCALES.keys()) adds all the scales from eessi.testsuite.constants as valid for thi partition
                    # Can be modified if not all scales can run on this partition, see e.g. the surf_snellius.py config
                    'features': [FEATURES[CPU]] + list(SCALES.keys()),
                },
                {
                    'name': 'gpu_partition',
                    'descr': 'GPU partition',
                    'scheduler': 'slurm',
                    'launcher': 'mpirun',
                    'access': ['-p gpu', '--export=None'],
                    'prepare_cmds': [common_eessi_init()],
                    'environs': ['default'],
                    'max_jobs': 4,
                    # We recommend to rely on ReFrame's CPU autodetection,
                    # and only define the 'processor' field if autodetection fails
                    # 'processor': {
                    #     'num_cpus': 72,
                    #     'num_sockets': 2,
                    #     'num_cpus_per_socket': 36,
                    #     'num_cpus_per_core': 1,
                    # },
                    'resources': [
                        {
                            'name': '_rfm_gpu',
                            'options': ['--gpus-per-node={num_gpus_per_node}'],
                        },
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'devices': [
                        {
                            'type': DEVICE_TYPES[GPU],
                            'num_devices': 4,
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                        FEATURES[GPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # If you have slurm, check with scontrol show node <nodename> for the amount of RealMemory
                        # on nodes in this partition
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 229376,  # in MiB
                        GPU_VENDOR: GPU_VENDORS[NVIDIA],
                    },
                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config()
        }
    ],
}



import tensorflow as tf
import numpy as np


def mnist_dataset(batch_size, test_batch_size):
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
    # The `x` arrays are in uint8 and have values in the [0, 255] range.
    # You need to convert them to float32 with values in the [0, 1] range.
    x_train = x_train / np.float32(255)
    y_train = y_train.astype(np.int64)
    x_test = x_test / np.float32(255)
    y_test = y_test.astype(np.int64)
    train_dataset = tf.data.Dataset.from_tensor_slices(
        (x_train, y_train)).shuffle(60000).repeat().batch(batch_size)
    test_dataset = tf.data.Dataset.from_tensor_slices(
        (x_test, y_test)).batch(test_batch_size)
    return train_dataset, test_dataset


def build_and_compile_cnn_model():
    model = tf.keras.Sequential([
        tf.keras.layers.InputLayer(input_shape=(28, 28)),
        tf.keras.layers.Reshape(target_shape=(28, 28, 1)),
        tf.keras.layers.Conv2D(32, 3, activation='relu'),
        tf.keras.layers.Flatten(),
        tf.keras.layers.Dense(128, activation='relu'),
        tf.keras.layers.Dense(10)
    ])
    model.compile(
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        metrics=['accuracy'])
    return model



"""
Hooks for adding tags, filtering and setting job resources in ReFrame tests
"""
import math

import reframe as rfm
import reframe.core.logging as rflog
import reframe.utility.sanity as sn

from eessi.testsuite.constants import (ALWAYS_REQUEST_GPUS, COMPUTE_UNIT, CPU, CPU_SOCKET, DEVICE_TYPES, FEATURES, GPU,
                                       GPU_VENDOR, GPU_VENDORS, HWTHREAD, INVALID_SYSTEM, NODE, NUMA_NODE, NVIDIA,
                                       SCALES)
from eessi.testsuite.utils import (get_max_avail_gpus_per_node, is_cuda_required_module, log,
                                   check_proc_attribute_defined, check_extras_key_defined)


def _assign_default_num_cpus_per_node(test: rfm.RegressionTest):
    """
    Check if the default number of cpus per node is already defined in the test
    (e.g. by earlier hooks like set_tag_scale).
    If so, check if it doesn't exceed the maximum available.
    If not, set default_num_cpus_per_node based on the maximum available cpus and node_part
    """

    if test.default_num_cpus_per_node:
        # may skip if not enough CPUs
        test.skip_if(
            test.default_num_cpus_per_node > test.max_avail_cpus_per_node,
            f'Requested CPUs per node ({test.default_num_cpus_per_node}) is higher than max available'
            f' ({test.max_avail_cpus_per_node}) in current partition ({test.current_partition.name}).'
        )
    else:
        # no default set yet, so setting one
        test.default_num_cpus_per_node = int(test.max_avail_cpus_per_node / test.node_part)

    log(f'default_num_cpus_per_node set to {test.default_num_cpus_per_node}')


def _assign_default_num_gpus_per_node(test: rfm.RegressionTest):
    """
    Check if the default number of gpus per node is already defined in the test
    (e.g. by earlier hooks like set_tag_scale).
    If so, check if it doesn't exceed the maximum available.
    If not, set default_num_gpus_per_node based on the maximum available gpus and node_part
    """

    test.max_avail_gpus_per_node = get_max_avail_gpus_per_node(test)
    if test.default_num_gpus_per_node:
        # may skip if not enough GPUs
        test.skip_if(
            test.default_num_gpus_per_node > test.max_avail_gpus_per_node,
            f'Number of GPUs per node in selected scale ({test.default_num_gpus_per_node}) is higher than max available'
            f' ({test.max_avail_gpus_per_node}) in current partition ({test.current_partition.name}).'
        )
    else:
        # no default set yet, so setting one
        test.default_num_gpus_per_node = math.ceil(test.max_avail_gpus_per_node / test.node_part)

    log(f'default_num_gpus_per_node set to {test.default_num_gpus_per_node}')


def assign_tasks_per_compute_unit(test: rfm.RegressionTest, compute_unit: str, num_per: int = 1):
    """
    Assign one task per compute unit. More than 1 task per compute unit can be assigned with
    num_per for compute units that support it.
    Automatically sets num_tasks, num_tasks_per_node, num_cpus_per_task, and num_gpus_per_node,
    based on the current scale and the current partition’s num_cpus, max_avail_gpus_per_node and num_nodes.
    For GPU tests, one task per GPU is set, and num_cpus_per_task is based on the ratio of CPU-cores/GPUs.
    For CPU tests, one task per CPU is set, and num_cpus_per_task is set to 1.
    Total task count is determined based on the number of nodes to be used in the test.
    Behaviour of this function is (usually) sensible for MPI tests.

    WARNING: when using COMPUTE_UNIT[HWTHREAD] and invoking a hook for process binding, please verify that process
    binding happens correctly.

    Arguments:
    - test: the ReFrame test to which this hook should apply
    - compute_unit: a device as listed in eessi.testsuite.constants.COMPUTE_UNIT

    Examples:
    On a single node with 2 sockets, 64 cores and 128 hyperthreads:
    - assign_tasks_per_compute_unit(test, COMPUTE_UNIT[HWTHREAD]) will launch 128 tasks with 1 thread per task
    - assign_tasks_per_compute_unit(test, COMPUTE_UNIT[CPU]) will launch 64 tasks with 2 threads per task
    - assign_tasks_per_compute_unit(test, COMPUTE_UNIT[CPU_SOCKET]) will launch 2 tasks with 64 threads per task

    """
    log(f'assign_tasks_per_compute_unit called with compute_unit: {compute_unit} and num_per: {num_per}')

    if num_per != 1 and compute_unit not in [COMPUTE_UNIT[NODE]]:
        raise NotImplementedError(
            f'Non-default num_per {num_per} is not implemented for compute_unit {compute_unit}.')

    check_proc_attribute_defined(test, 'num_cpus')
    test.max_avail_cpus_per_node = test.current_partition.processor.num_cpus
    log(f'max_avail_cpus_per_node set to {test.max_avail_cpus_per_node}')

    # Check if either node_part, or default_num_cpus_per_node and default_num_gpus_per_node are set correctly
    if not (
        isinstance(test.node_part, int)
        or (isinstance(test.default_num_cpus_per_node, int) and isinstance(test.default_num_gpus_per_node, int))
    ):
        raise ValueError(
            f'Either node_part ({test.node_part}), or default_num_cpus_per_node ({test.default_num_cpus_per_node}) and'
            f' default num_gpus_per_node ({test.default_num_gpus_per_node}) must be defined and have integer values.'
        )

    _assign_default_num_cpus_per_node(test)
    # If on
    # - a hyperthreading system
    # - num_cpus_per_node was set by the scale
    # - compute_unit != COMPUTE_UNIT[HWTHREAD]
    # double the default_num_cpus_per_node. In this scenario, if the scale asks for e.g. 1 num_cpus_per_node and
    # the test doesn't state it wants to use hwthreads, we want to launch on two hyperthreads, i.e. one physical core
    if SCALES[test.scale].get('num_cpus_per_node') is not None and compute_unit != COMPUTE_UNIT[HWTHREAD]:
        check_proc_attribute_defined(test, 'num_cpus_per_core')
        num_cpus_per_core = test.current_partition.processor.num_cpus_per_core
        # On a hyperthreading system?
        if num_cpus_per_core > 1:
            test.default_num_cpus_per_node = test.default_num_cpus_per_node * num_cpus_per_core

    if FEATURES[GPU] in test.current_partition.features:
        _assign_default_num_gpus_per_node(test)

    if compute_unit == COMPUTE_UNIT[GPU]:
        _assign_one_task_per_gpu(test)
    elif compute_unit == COMPUTE_UNIT[HWTHREAD]:
        _assign_one_task_per_hwthread(test)
    elif compute_unit == COMPUTE_UNIT[CPU]:
        _assign_one_task_per_cpu(test)
    elif compute_unit == COMPUTE_UNIT[CPU_SOCKET]:
        _assign_one_task_per_cpu_socket(test)
    elif compute_unit == COMPUTE_UNIT[NUMA_NODE]:
        _assign_one_task_per_numa_node(test)
    elif compute_unit == COMPUTE_UNIT[NODE]:
        _assign_num_tasks_per_node(test, num_per)
    else:
        raise ValueError(f'compute unit {compute_unit} is currently not supported')

    _check_always_request_gpus(test)

    if test.current_partition.launcher_type().registered_name == 'srun':
        # Make sure srun inherits --cpus-per-task from the job environment for Slurm versions >= 22.05 < 23.11,
        # ensuring the same task binding across all Slurm versions.
        # https://bugs.schedmd.com/show_bug.cgi?id=13351
        # https://bugs.schedmd.com/show_bug.cgi?id=11275
        # https://bugs.schedmd.com/show_bug.cgi?id=15632#c43
        test.env_vars['SRUN_CPUS_PER_TASK'] = test.num_cpus_per_task
        log(f'Set environment variable SRUN_CPUS_PER_TASK to {test.env_vars["SRUN_CPUS_PER_TASK"]}')


def _assign_num_tasks_per_node(test: rfm.RegressionTest, num_per: int = 1):
    """
    Sets num_tasks_per_node and num_cpus_per_task such that it will run
    'num_per' tasks per node, unless specified with:
    --setvar num_tasks_per_node=<x>
    --setvar num_cpus_per_task=<y>.
    In those cases, those take precedence, and the remaining variable, if any
    (num_cpus_per task or num_tasks_per_node respectively), is calculated based
    on the equality test.num_tasks_per_node * test.num_cpus_per_task ==
    test.default_num_cpus_per_node.

    Default resources requested:
    - num_tasks_per_node = num_per
    - num_cpus_per_task = test.default_num_cpus_per_node / num_tasks_per_node
    """

    # neither num_tasks_per_node nor num_cpus_per_task are set
    if not test.num_tasks_per_node and not test.num_cpus_per_task:
        test.num_tasks_per_node = num_per
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    # num_tasks_per_node is not set, but num_cpus_per_task is
    elif not test.num_tasks_per_node:
        test.num_tasks_per_node = int(test.default_num_cpus_per_node / test.num_cpus_per_task)

    # num_cpus_per_task is not set, but num_tasks_per_node is
    elif not test.num_cpus_per_task:
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    else:
        pass  # both num_tasks_per_node and num_cpus_per_task are already set

    test.num_tasks = test.num_nodes * test.num_tasks_per_node

    log(f'num_tasks_per_node set to {test.num_tasks_per_node}')
    log(f'num_cpus_per_task set to {test.num_cpus_per_task}')
    log(f'num_tasks set to {test.num_tasks}')


def _assign_one_task_per_cpu_socket(test: rfm.RegressionTest):
    """
    Determines the number of tasks per node by dividing the default_num_cpus_per_node by
    the number of cpus available per socket, and rounding up. The result is that for full-node jobs the default
    will spawn one task per socket, with a number of cpus per task equal to the number of cpus per socket.
    Other examples:
    - half a node (i.e. node_part=2) on a 4-socket system would result in 2 tasks per node,
    with number of cpus per task equal to the number of cpus per socket.
    - 2 cores (i.e. default_num_cpus_per_node=2) on a 16 core system with 2 sockets would result in
    1 task per node, with 2 cpus per task

    This default is set unless the test is run with:
    --setvar num_tasks_per_node=<x> and/or
    --setvar num_cpus_per_task=<y>.
    In those cases, those take precedence, and the remaining variable (num_cpus_per task or
    num_tasks_per_node respectively) is calculated based on the equality
    test.num_tasks_per_node * test.num_cpus_per_task == test.default_num_cpus_per_node.

    Default resources requested:
    - num_tasks_per_node = default_num_cpus_per_node / num_cpus_per_socket
    - num_cpus_per_task = default_num_cpus_per_node / num_tasks_per_node
    """
    # neither num_tasks_per_node nor num_cpus_per_task are set
    if not test.num_tasks_per_node and not test.num_cpus_per_task:
        check_proc_attribute_defined(test, 'num_cores_per_socket')
        test.num_tasks_per_node = math.ceil(
            test.default_num_cpus_per_node / test.current_partition.processor.num_cores_per_socket
        )
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    # num_tasks_per_node is not set, but num_cpus_per_task is
    elif not test.num_tasks_per_node:
        test.num_tasks_per_node = int(test.default_num_cpus_per_node / test.num_cpus_per_task)

    # num_cpus_per_task is not set, but num_tasks_per_node is
    elif not test.num_cpus_per_task:
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    else:
        pass  # both num_tasks_per_node and num_cpus_per_node are already set

    test.num_tasks = test.num_nodes * test.num_tasks_per_node
    log(f'Number of tasks per node set to: {test.num_tasks_per_node}')
    log(f'Number of cpus per task set to {test.num_cpus_per_task}')
    log(f'num_tasks set to {test.num_tasks}')


def _assign_one_task_per_numa_node(test: rfm.RegressionTest):
    """
    Determines the number of tasks per node by dividing the default_num_cpus_per_node by
    the number of cpus available per numa node, and rounding up. The result is that for full-node jobs the default
    will spawn one task per numa node, with a number of cpus per task equal to the number of cpus per numa node.
    Other examples:
    - half a node (i.e. node_part=2) on a system with 4 numa nodes would result in 2 tasks per node,
    with number of cpus per task equal to the number of cpus per numa node.
    - a quarter node (i.e. node_part=4) on a system with 2 numa nodes would result in 1 task per node,
    with number of cpus equal to half a numa node.
    - 2 cores (i.e. default_num_cpus_per_node=2) on a system with 4 cores per numa node would result in
    1 task per node, with 2 cpus per task
    - 8 cores (i.e. default_num_cpus_per_node=4) on a system with 4 cores per numa node would result in
    2 tasks per node, with 4 cpus per task

    This default is set unless the test is run with:
    --setvar num_tasks_per_node=<x> and/or
    --setvar num_cpus_per_task=<y>.
    In those cases, those take precedence, and the remaining variable (num_cpus_per task or
    num_tasks_per_node respectively) is calculated based on the equality
    test.num_tasks_per_node * test.num_cpus_per_task == test.default_num_cpus_per_node.

    Default resources requested:
    - num_tasks_per_node = default_num_cpus_per_node / num_cores_per_numa_node
    - num_cpus_per_task = default_num_cpus_per_node / num_tasks_per_node
    """
    # neither num_tasks_per_node nor num_cpus_per_task are set
    if not test.num_tasks_per_node and not test.num_cpus_per_task:
        check_proc_attribute_defined(test, 'num_cores_per_numa_node')
        test.num_tasks_per_node = math.ceil(
            test.default_num_cpus_per_node / test.current_partition.processor.num_cores_per_numa_node
        )
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    # num_tasks_per_node is not set, but num_cpus_per_task is
    elif not test.num_tasks_per_node:
        test.num_tasks_per_node = int(test.default_num_cpus_per_node / test.num_cpus_per_task)

    # num_cpus_per_task is not set, but num_tasks_per_node is
    elif not test.num_cpus_per_task:
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    else:
        pass  # both num_tasks_per_node and num_cpus_per_node are already set

    test.num_tasks = test.num_nodes * test.num_tasks_per_node
    log(f'Number of tasks per node set to: {test.num_tasks_per_node}')
    log(f'Number of cpus per task set to {test.num_cpus_per_task}')
    log(f'num_tasks set to {test.num_tasks}')


def _assign_one_task_per_cpu(test: rfm.RegressionTest):
    """
    Sets num_tasks_per_node and num_cpus_per_task such that it will run one task per core,
    unless specified with:
    --setvar num_tasks_per_node=<x> and/or
    --setvar num_cpus_per_task=<y>.

    Default resources requested:
    - num_tasks_per_node = default_num_cpus_per_node
    - num_cpus_per_task = default_num_cpus_per_node / num_tasks_per_node
    """
    # neither num_tasks_per_node nor num_cpus_per_task are set
    if not test.num_tasks_per_node and not test.num_cpus_per_task:
        check_proc_attribute_defined(test, 'num_cpus_per_core')
        test.num_tasks_per_node = max(
            int(test.default_num_cpus_per_node / test.current_partition.processor.num_cpus_per_core),
            1
        )
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    # num_tasks_per_node is not set, but num_cpus_per_task is
    elif not test.num_tasks_per_node:
        test.num_tasks_per_node = int(test.default_num_cpus_per_node / test.num_cpus_per_task)

    # num_cpus_per_task is not set, but num_tasks_per_node is
    elif not test.num_cpus_per_task:
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    else:
        pass  # both num_tasks_per_node and num_cpus_per_node are already set

    test.num_tasks = test.num_nodes * test.num_tasks_per_node

    log(f'num_tasks_per_node set to {test.num_tasks_per_node}')
    log(f'num_cpus_per_task set to {test.num_cpus_per_task}')
    log(f'num_tasks set to {test.num_tasks}')


def _assign_one_task_per_hwthread(test: rfm.RegressionTest):
    """
    Sets num_tasks_per_node and num_cpus_per_task such that it will run one task per core,
    unless specified with:
    --setvar num_tasks_per_node=<x> and/or
    --setvar num_cpus_per_task=<y>.

    Default resources requested:
    - num_tasks_per_node = default_num_cpus_per_node
    - num_cpus_per_task = default_num_cpus_per_node / num_tasks_per_node
    """
    # neither num_tasks_per_node nor num_cpus_per_task are set
    if not test.num_tasks_per_node and not test.num_cpus_per_task:
        test.num_tasks_per_node = test.default_num_cpus_per_node
        test.num_cpus_per_task = 1

    # num_tasks_per_node is not set, but num_cpus_per_task is
    elif not test.num_tasks_per_node:
        test.num_tasks_per_node = int(test.default_num_cpus_per_node / test.num_cpus_per_task)

    # num_cpus_per_task is not set, but num_tasks_per_node is
    elif not test.num_cpus_per_task:
        test.num_cpus_per_task = int(test.default_num_cpus_per_node / test.num_tasks_per_node)

    else:
        pass  # both num_tasks_per_node and num_cpus_per_node are already set

    test.num_tasks = test.num_nodes * test.num_tasks_per_node

    log(f'num_tasks_per_node set to {test.num_tasks_per_node}')
    log(f'num_cpus_per_task set to {test.num_cpus_per_task}')
    log(f'num_tasks set to {test.num_tasks}')


def _assign_one_task_per_gpu(test: rfm.RegressionTest):
    """
    Sets num_tasks_per_node, num_cpus_per_task, and num_gpus_per_node, unless specified with:
    --setvar num_tasks_per_node=<x> and/or
    --setvar num_cpus_per_task=<y> and/or
    --setvar num_gpus_per_node=<z>.

    Default resources requested:
    - num_gpus_per_node = default_num_gpus_per_node
    - num_tasks_per_node = num_gpus_per_node
    - num_cpus_per_task = default_num_cpus_per_node / num_tasks_per_node

    If num_tasks_per_node is set, set num_gpus_per_node equal to either num_tasks_per_node or default_num_gpus_per_node
    (whichever is smallest), unless num_gpus_per_node is also set.
    """

    # neither num_tasks_per_node nor num_gpus_per_node are set
    if not test.num_tasks_per_node and not test.num_gpus_per_node:
        test.num_gpus_per_node = test.default_num_gpus_per_node
        test.num_tasks_per_node = test.num_gpus_per_node

    # num_tasks_per_node is not set, but num_gpus_per_node is
    elif not test.num_tasks_per_node:
        test.num_tasks_per_node = test.num_gpus_per_node

    # num_gpus_per_node is not set, but num_tasks_per_node is
    elif not test.num_gpus_per_node:
        test.num_gpus_per_node = min(test.num_tasks_per_node, test.default_num_gpus_per_node)

    else:
        pass  # both num_tasks_per_node and num_gpus_per_node are already set

    # num_cpus_per_task is not set
    if not test.num_cpus_per_task:
        # limit num_cpus_per_task to the maximum available cpus per gpu
        test.num_cpus_per_task = min(
            int(test.default_num_cpus_per_node / test.num_tasks_per_node),
            int(test.max_avail_cpus_per_node / test.max_avail_gpus_per_node)
        )

    test.num_tasks = test.num_nodes * test.num_tasks_per_node

    log(f'num_gpus_per_node set to {test.num_gpus_per_node}')
    log(f'num_tasks_per_node set to {test.num_tasks_per_node}')
    log(f'num_cpus_per_task set to {test.num_cpus_per_task}')
    log(f'num_tasks set to {test.num_tasks}')


def _set_or_append_valid_systems(test: rfm.RegressionTest, valid_systems: str):
    """
    Sets test.valid_systems based on the valid_systems argument.
    - If valid_systems is an empty string, test.valid_systems is set equal to eessi.testsuite.constants.INVALID_SYSTEM
    - If test.valid_systems was an empty list, leave it as is (test should not be run)
    - If test.valid_systems was at the default value ['*'], it is overwritten by [valid_system]
    - If test.valid_systems was already set and is a list of one element, valid_system is appended to it,
    which allows adding requests for multiple partition features by different hooks.
    - If test.valid_systems was already set and is a list of multiple elements, we warn that the use has to take
    care of filtering him/herself. This is typically the case when someone overrides the valid_systems on command line.
    In this scenario, this function leaves test.valid_systems as it is.
    """

    # This indicates an invalid test that always has to be filtered
    if valid_systems == '':
        test.valid_systems = [INVALID_SYSTEM]
        return

    # test.valid_systems wasn't set yet, so set it
    if len(test.valid_systems) == 0 or test.valid_systems == [INVALID_SYSTEM]:
        # test.valid_systems is empty or invalid, meaning all tests are filtered out. This hook shouldn't change that
        return
    # test.valid_systems still at default value, so overwrite
    elif len(test.valid_systems) == 1 and test.valid_systems[0] == '*':
        test.valid_systems = [valid_systems]
    # test.valid_systems was set before, so append
    elif len(test.valid_systems) == 1:
        test.valid_systems[0] = f'{test.valid_systems[0]} {valid_systems}'
    else:
        msg = f"valid_systems has multiple ({len(test.valid_systems)}) items,"
        msg += " which is not supported by this hook."
        msg += " Make sure to handle filtering yourself."
        rflog.getlogger().warning(msg)
        return


def filter_supported_scales(test: rfm.RegressionTest):
    """
    Filter tests scales based on which scales are supported by each partition in the ReFrame configuration.
    Filtering is done using features, i.e. the current test scale is requested as a feature.
    Any partition that does not include this feature in the ReFrame configuration file will effectively be filtered out.
    """
    valid_systems = f'+{test.scale}'

    # Change test.valid_systems accordingly:
    _set_or_append_valid_systems(test, valid_systems)

    log(f'valid_systems set to {test.valid_systems}')


def filter_valid_systems_by_device_type(test: rfm.RegressionTest, required_device_type: str):
    """
    Filter valid_systems by required device type and by whether the module supports CUDA,
    unless valid_systems is specified with --setvar valid_systems=<comma-separated-list>.

    Any invalid combination (e.g. a non-CUDA module with a required_device_type GPU) will
    cause the valid_systems to be set to an empty string, and consequently the
    test.valid_systems to an invalid system name (eessi.testsuite.constants.INVALID_SYSTEM).
    """
    is_cuda_module = is_cuda_required_module(test.module_name)

    if is_cuda_module and required_device_type == DEVICE_TYPES[GPU]:
        # CUDA modules and when using a GPU require partitions with FEATURES[GPU] feature and
        # GPU_VENDOR=GPU_VENDORS[NVIDIA] extras
        valid_systems = f'+{FEATURES[GPU]} %{GPU_VENDOR}={GPU_VENDORS[NVIDIA]}'

    elif not is_cuda_module and required_device_type == DEVICE_TYPES[CPU]:
        # Using the CPU requires partitions with FEATURES[CPU] feature
        # Note: making FEATURES[CPU] an explicit feature allows e.g. skipping CPU-based tests on GPU partitions
        valid_systems = f'+{FEATURES[CPU]}'

    elif is_cuda_module and required_device_type == DEVICE_TYPES[CPU]:
        # Note: This applies for CUDA module tests that want to test only on cpus on gpu partitions.
        valid_systems = f'+{FEATURES[CPU]} +{FEATURES[GPU]} %{GPU_VENDOR}={GPU_VENDORS[NVIDIA]}'

    elif not is_cuda_module and required_device_type == DEVICE_TYPES[GPU]:
        # Invalid combination: a module without GPU support cannot use a GPU
        valid_systems = ''

    # Change test.valid_systems accordingly:
    _set_or_append_valid_systems(test, valid_systems)

    log(f'valid_systems set to {test.valid_systems}')


def req_memory_per_node(test: rfm.RegressionTest, app_mem_req: float):
    """
    This hook will request a specific amount of memory per node to the batch scheduler.
    First, it computes which fraction of CPUs is requested from a node, and how much the corresponding (proportional)
    amount of memory would be.
    Then, the hook compares this to how much memory the application claims to need per node (app_mem_req).
    It then passes the maximum of these two numbers to the batch scheduler as a memory request.

    Note: using this hook requires that the ReFrame configuration defines system.partition.extras['mem_per_node']
    That field should be defined in GiB

    Arguments:
    - test: the ReFrame test to which this hook should apply
    - app_mem_req: the amount of memory this application needs (per node) in MiB

    Example 1:
    - A system with 128 cores and 64,000 MiB per node.
    - The test is launched on 64 cores
    - The app_mem_req is 40,000 (MiB)
    In this case, the test requests 50% of the CPUs. Thus, the proportional amount of memory is 32,000 MiB.
    The app_mem_req is higher. Thus, 40,000 MiB (per node) is requested from the batch scheduler.

    Example 2:
    - A system with 128 cores per node, 128,000 MiB mem per node.
    - The test is launched on 64 cores
    - the app_mem_req is 40,000 (MiB)
    In this case, the test requests 50% of the CPUs. Thus, the proportional amount of memory is 64,000 MiB.
    This is higher than the app_mem_req. Thus, 64,000 MiB (per node) is requested from the batch scheduler.
    """
    # Check that the systems.partitions.extra dict in the ReFrame config contains mem_per_node
    check_extras_key_defined(test, 'mem_per_node')
    # Skip if the current partition doesn't have sufficient memory to run the application
    msg = f"Skipping test: nodes in this partition only have {test.current_partition.extras['mem_per_node']} MiB"
    msg += " memory available (per node) accodring to the current ReFrame configuration,"
    msg += f" but {app_mem_req} MiB is needed"
    test.skip_if(test.current_partition.extras['mem_per_node'] < app_mem_req, msg)

    # Check if a resource with the name 'memory' was set in the ReFrame config file. If not, warn the user
    # and return from this hook (as setting test.extra_resources will be ignored in that case according to
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.extra_resources
    if 'memory' not in test.current_partition.resources:
        msg = "Your ReFrame configuration file does not specify any resource called 'memory' for this partition "
        msg += f" ({test.current_partition.name})."
        msg += " Without this, an explicit memory request cannot be made from the scheduler. This test will run,"
        msg += " but it may result in an out of memory error."
        msg += " Please specify how to requst memory (per node) from your resource scheduler by defining a resource"
        msg += " 'memory' in your ReFrame configuration file for this partition."
        msg += " For a SLURM system, one would e.g. define:"
        msg += " 'resources': [{'name': 'memory', 'options': ['--mem={size}']}]"
        rflog.getlogger().warning(msg)
        # We return, as setting a test.extra_resources is pointless - it would be ignored anyway
        # This way, we also don't add any lines to the log that a specific amount of memory was requested
        return

    # Compute what is higher: the requested memory, or the memory available proportional to requested CPUs
    # Fraction of CPU cores requested
    check_proc_attribute_defined(test, 'num_cpus')
    cpu_fraction = test.num_tasks_per_node * test.num_cpus_per_task / test.current_partition.processor.num_cpus
    proportional_mem = math.floor(cpu_fraction * test.current_partition.extras['mem_per_node'])
    app_mem_req = math.ceil(app_mem_req)

    scheduler_name = test.current_partition.scheduler.registered_name
    if scheduler_name == 'slurm' or scheduler_name == 'squeue':
        # SLURM defines --mem as memory per node, see https://slurm.schedmd.com/sbatch.html
        # SLURM uses MiB units by default
        log(f"Memory requested by application: {app_mem_req} MiB")
        log(f"Memory proportional to the core count: {proportional_mem} MiB")

        # make sure exact_memory is always defined for tests that don't (yet) use the EESSI_Mixin class
        if not hasattr(test, 'exact_memory'):
            test.exact_memory = False

        if test.exact_memory:
            # Request the exact amount of required memory
            req_mem_per_node = app_mem_req
        else:
            # Request the maximum of the proportional_mem, and app_mem_req to the scheduler
            req_mem_per_node = max(proportional_mem, app_mem_req)

        test.extra_resources = {'memory': {'size': f'{req_mem_per_node}M'}}
        log(f"Requested {req_mem_per_node} MiB per node from the SLURM batch scheduler")

    elif scheduler_name == 'torque':
        # Torque/moab requires asking for --pmem (--mem only works single node and thus doesnt generalize)
        # See https://docs.adaptivecomputing.com/10-0-1/Torque/torque.htm#topics/torque/3-jobs/3.1.3-requestingRes.htm
        # Units are MiB according to the documentation
        # We immediately divide by num_tasks_per_node (before rounding), since -pmem specifies memroy _per process_
        app_mem_req_task = math.ceil(app_mem_req / test.num_tasks_per_node)
        proportional_mem_task = math.floor(proportional_mem / test.num_tasks_per_node)

        # Request the maximum of the proportional_mem, and app_mem_req to the scheduler
        req_mem_per_task = max(proportional_mem_task, app_mem_req_task)

        # We assume here the reframe config defines the extra resource memory as asking for pmem
        # i.e. 'options': ['--pmem={size}']
        test.extra_resources = {'memory': {'size': f'{req_mem_per_task}mb'}}
        log(f"Requested {req_mem_per_task} MiB per task from the torque batch scheduler")

    else:
        msg = "hooks.req_memory_per_node does not support the scheduler you configured"
        msg += f" ({test.current_partition.scheduler.registered_name})."
        msg += " The test will run, but since it doesn't request the required amount of memory explicitely,"
        msg += " it may result in an out-of-memory error."
        msg += " Please expand the functionality of hooks.req_memory_per_node for your scheduler."
        # Warnings will, at default loglevel, be printed on stdout when executing the ReFrame command
        rflog.getlogger().warning(msg)


def set_modules(test: rfm.RegressionTest):
    """
    Skip current test if module_name is not among a list of modules,
    specified with --setvar modules=<comma-separated-list>.
    """
    if test.modules and test.module_name not in test.modules:
        test.valid_systems = []
        log(f'valid_systems set to {test.valid_systems}')

    test.modules = [test.module_name]
    log(f'modules set to {test.modules}')


def set_tag_scale(test: rfm.RegressionTest):
    """Set resources and tag based on current scale"""
    scale = test.scale
    test.num_nodes = SCALES[scale]['num_nodes']
    test.default_num_cpus_per_node = SCALES[scale].get('num_cpus_per_node')
    test.default_num_gpus_per_node = SCALES[scale].get('num_gpus_per_node')
    test.node_part = SCALES[scale].get('node_part')
    test.tags.add(scale)
    log(f'tags set to {test.tags}')


def set_compact_process_binding(test: rfm.RegressionTest):
    """
    This hook sets a binding policy for process binding.
    More specifically, it will bind each process to subsequent domains of test.num_cpus_per_task cores.

    Arguments:
    - test: the ReFrame test to which this hook should apply


    A few examples:
    - Pure MPI (test.num_cpus_per_task = 1) will result in binding 1 process to each core.
      this will happen in a compact way, i.e. rank 0 to core 0, rank 1 to core 1, etc
    - Hybrid MPI-OpenMP, e.g. test.num_cpus_per_task = 4 will result in binding 1 process to subsequent sets of 4 cores.
      I.e. rank 0 to core 0-3, rank 1 to core 4-7, rank 2 to core 8-11, etc

    It is hard to do this in a portable way. Currently supported for process binding are:
    - Intel MPI (through I_MPI_PIN_DOMAIN)
    - OpenMPI (through OMPI_MCA_rmaps_base_mapping_policy)
    - srun (LIMITED SUPPORT: through SLURM_CPU_BIND, but only effective if task/affinity plugin is enabled)
    """

    # Check if hyperthreading is enabled. If so, divide the number of cpus per task by the number
    # of hw threads per core to get a physical core count
    # TODO: check if this also leads to sensible binding when using COMPUTE_UNIT[HWTHREAD]
    check_proc_attribute_defined(test, 'num_cpus_per_core')
    num_cpus_per_core = test.current_partition.processor.num_cpus_per_core
    physical_cpus_per_task = int(test.num_cpus_per_task / num_cpus_per_core)

    if test.current_partition.launcher_type().registered_name == 'mpirun':
        # Do binding for intel and OpenMPI's mpirun, and srun
        test.env_vars['I_MPI_PIN_CELL'] = 'core'  # Don't bind to hyperthreads, only to physcial cores
        test.env_vars['I_MPI_PIN_DOMAIN'] = '%s:compact' % physical_cpus_per_task
        test.env_vars['OMPI_MCA_rmaps_base_mapping_policy'] = 'slot:PE=%s' % physical_cpus_per_task
        log(f'Set environment variable I_MPI_PIN_CELL to {test.env_vars["I_MPI_PIN_CELL"]}')
        log(f'Set environment variable I_MPI_PIN_DOMAIN to {test.env_vars["I_MPI_PIN_DOMAIN"]}')
        log('Set environment variable OMPI_MCA_rmaps_base_mapping_policy to '
            f'{test.env_vars["OMPI_MCA_rmaps_base_mapping_policy"]}')
    elif test.current_partition.launcher_type().registered_name == 'srun':
        # Set compact binding for SLURM. Only effective if the task/affinity plugin is enabled
        # and when number of tasks times cpus per task equals either socket, core or thread count
        test.env_vars['SLURM_DISTRIBUTION'] = 'block:block'
        test.env_vars['SLURM_CPU_BIND'] = 'verbose'
        log(f'Set environment variable SLURM_DISTRIBUTION to {test.env_vars["SLURM_DISTRIBUTION"]}')
        log(f'Set environment variable SLURM_CPU_BIND to {test.env_vars["SLURM_CPU_BIND"]}')
    else:
        msg = "hooks.set_compact_process_binding does not support the current launcher"
        msg += f" ({test.current_partition.launcher_type().registered_name})."
        msg += " The test will run, but using the default binding strategy of your parallel launcher."
        msg += " This may lead to suboptimal performance."
        msg += " Please expand the functionality of hooks.set_compact_process_binding for your parallel launcher."
        # Warnings will, at default loglevel, be printed on stdout when executing the ReFrame command
        rflog.getlogger().warning(msg)


def set_compact_thread_binding(test: rfm.RegressionTest):
    """
    This hook sets a binding policy for thread binding.
    It sets a number of environment variables to try and set a sensible binding for OPENMP tasks.

    Thread binding is supported for:
    - GNU OpenMP (through OMP_NUM_THREADS, OMP_PLACES and OMP_PROC_BIND)
    - Intel OpenMP (through KMP_AFFINITY)
    """

    # Set thread binding
    test.env_vars['OMP_PLACES'] = 'cores'
    test.env_vars['OMP_PROC_BIND'] = 'close'
    # See https://www.intel.com/content/www/us/en/docs/cpp-compiler/developer-guide-reference/2021-8/thread-affinity-interface.html  # noqa
    test.env_vars['KMP_AFFINITY'] = 'granularity=fine,compact,1,0'
    log(f'Set environment variable OMP_PLACES to {test.env_vars["OMP_PLACES"]}')
    log(f'Set environment variable OMP_PROC_BIND to {test.env_vars["OMP_PROC_BIND"]}')
    log(f'Set environment variable KMP_AFFINITY to {test.env_vars["KMP_AFFINITY"]}')


def set_omp_num_threads(test: rfm.RegressionTest):
    """
    Set number of OpenMP threads equal to number of CPUs per task
    """
    test.env_vars['OMP_NUM_THREADS'] = test.num_cpus_per_task
    log(f'Set environment variable OMP_NUM_THREADS to {test.env_vars["OMP_NUM_THREADS"]}')


def _check_always_request_gpus(test: rfm.RegressionTest):
    """
    Make sure we always request enough GPUs if required for the current GPU partition (cluster-specific policy)
    """
    # make sure always_request_gpus is always defined for tests that don't (yet) use the EESSI_Mixin class
    if not hasattr(test, 'always_request_gpus'):
        test.always_request_gpus = False

    always_request_gpus = FEATURES[ALWAYS_REQUEST_GPUS] in test.current_partition.features or test.always_request_gpus
    if always_request_gpus and not test.num_gpus_per_node:
        test.num_gpus_per_node = test.default_num_gpus_per_node
        log(f'num_gpus_per_node set to {test.num_gpus_per_node} for partition {test.current_partition.name}')


def measure_memory_usage(test: rfm.RegressionTest):
    """
    Write the memory usage into the job output file if we are in a Slurm job and if cgroups is enabled in Slurm
    First try to obtain the memory with cgroups v2, if that fails try with cgroups v1 (v2 takes precedence)
    Intended to be used in tandem with hook extract_memory_usage()
    To use this hook, add the following method to your test class:

    @run_after('init')
    def measure_memory_usage(self):
        "Measure memory usage"
        hooks.measure_memory_usage(self)
    """
    test.postrun_cmds = [
        'path_v2=/sys/fs/cgroup/$(</proc/self/cpuset)/../../../memory.peak',
        'path_v1=/sys/fs/cgroup/memory/$(</proc/self/cpuset)/../memory.max_usage_in_bytes',
        'if [[ -f $path_v2 ]]; then MAX_MEM_IN_BYTES=$(<$path_v2)',
        'elif [[ -f $path_v1 ]]; then MAX_MEM_IN_BYTES=$(<$path_v1)',
        'fi',
        '[[ $MAX_MEM_IN_BYTES =~ ^-?[0-9]+$ ]] || (echo ERROR: unable to get memory usage; exit 1)',
        'echo "MAX_MEM_IN_BYTES=$MAX_MEM_IN_BYTES"',
        'echo "MAX_MEM_IN_MIB=$(($MAX_MEM_IN_BYTES/1048576))"',
    ]


def extract_memory_usage(test: rfm.RegressionTest):
    """
    Extract the memory in MiB from the job output file as written by hook measure_memory_usage()
    To Use this hook, add the following method to your test class:

    @performance_function('MiB', perf_key='memory')
    def extract_memory_usage(self):
        return hooks.extract_memory_usage(self)
    """
    return sn.extractsingle(r'^MAX_MEM_IN_MIB=(?P<memory>\S+)', test.stdout, 'memory', int)



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914

import os

from eessi.testsuite.common_config import common_logging_config, common_eessi_init
from eessi.testsuite.constants import FEATURES

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# AWS CITC site configuration
site_configuration = {
    'systems': [
        {
            'name': 'citc',
            'descr': 'Cluster in the Cloud build and test environment on AWS',
            'modules_system': 'lmod',
            'hostnames': ['mgmt', 'login', 'fair-mastodon*'],
            'prefix': reframe_prefix,
            'partitions': [
                {
                    'name': 'x86_64-haswell-8c-15gb',
                    'access': ['--constraint=shape=c4.2xlarge', '--export=NONE'],
                    'descr': 'Haswell, 8 cores, 15 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-haswell-16c-30gb',
                    'access': ['--constraint=shape=c4.4xlarge', '--export=NONE'],
                    'descr': 'Haswell, 16 cores, 30 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-zen2-8c-16gb',
                    'access': ['--constraint=shape=c5a.2xlarge', '--export=NONE'],
                    'descr': 'Zen2, 8 cores, 16 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-zen2-16c-32gb',
                    'access': ['--constraint=shape=c5a.4xlarge', '--export=NONE'],
                    'descr': 'Zen2, 16 cores, 32 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-zen3-8c-16gb',
                    'access': ['--constraint=shape=c6a.2xlarge', '--export=NONE'],
                    'descr': 'Zen3, 8 cores, 16 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'X86_64-zen3-16c-32gb',
                    'access': ['--constraint=shape=c6a.4xlarge', '--export=NONE'],
                    'descr': 'Zen3, 16 cores, 32 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-skylake-cascadelake-8c-16gb',
                    'access': ['--constraint=shape=c5.2xlarge', '--export=NONE'],
                    'descr': 'Skylake/Cascade lake, 8 cores, 16 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-skylake-cascadelake-16c-32gb',
                    'access': ['--constraint=shape=c5.4xlarge', '--export=NONE'],
                    'descr': 'Skylake/Cascade lake, 16 cores, 32 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-skylake-cascadelake-8c-16gb-nvme',
                    'access': ['--constraint=shape=c5d.2xlarge', '--export=NONE'],
                    'descr': 'Skylake/Cascade lake, 8 cores, 16 GiB, 200GB NVMe',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'x86_64-icelake-8c-16gb',
                    'access': ['--constraint=shape=c6i.2xlarge', '--export=NONE'],
                    'descr': 'Icelake, 8 cores, 16 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'aarch64-graviton2-8c-16gb',
                    'access': ['--constraint=shape=c6g.2xlarge', '--export=NONE'],
                    'descr': 'Graviton2, 8 cores, 16 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'aarch64-graviton2-16c-32gb',
                    'access': ['--constraint=shape=c6g.4xlarge', '--export=NONE'],
                    'descr': 'Graviton2, 16 cores, 32 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'aarch64-graviton2-32c-64gb',
                    'access': ['--constraint=shape=c6g.8xlarge', '--export=NONE'],
                    'descr': 'Graviton2, 32 cores, 64 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'aarch64-graviton3-8c-16gb',
                    'access': ['--constraint=shape=c7g.2xlarge', '--export=NONE'],
                    'descr': 'Graviton3, 8 cores, 16 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
                {
                    'name': 'aarch64-graviton3-16c-32gb',
                    'access': ['--constraint=shape=c7g.4xlarge', '--export=NONE'],
                    'descr': 'Graviton3, 16 cores, 32 GiB',
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
        }
    ],
}

# Add default things to each partition:
partition_defaults = {
    'scheduler': 'squeue',
    'launcher': 'mpirun',
    'environs': ['default'],
    'features': [
        FEATURES['CPU']
    ],
    'prepare_cmds': [
        common_eessi_init(),
        # Required when using srun as launcher with --export=NONE in partition access, in order to ensure job
        # steps inherit environment. It doesn't hurt to define this even if srun is not used
        'export SLURM_EXPORT_ENV=ALL'
    ],
}
for system in site_configuration['systems']:
    for partition in system['partitions']:
        partition.update(partition_defaults)



# WARNING: this file is imported in setup.py
# To make sure this works, we should avoid using imports other than from the Python standard library

try:
    # If this is an installed package, setuptools_scm will have written the _version.py file in the current directory
    from ._version import __version__
except ImportError:
    try:
        # Setuptools_scm 4.1.2 (compatible with setuptools 39.2.0) write version instead of __version__
        # This can be removed once we no longer care about python 3.6 with setuptools 39.2.0
        from ._version import version
        __version__ = version
    except ImportError:
        # Fallback for when the package is not installed, but git cloned. Note that this requires setuptools_scm to be
        # available as a runtime dependency
        # The advantage here is that it will generate development versions if not on a tagged release version
        try:
            from setuptools_scm import get_version
            # Using a relative path for relative_to doesn't work, because it will be relative to the current working
            # directory (which could be anywhere)
            # __file__ is the location of this init file (a full path), and this gives us a predictable path to the root
            # (namely: two levels up). Note that if we ever move this __init__ file relative to the root of the git
            # tree, we'll need to adjust this
            __version__ = get_version(root='../..', relative_to=__file__)
        except (ImportError, LookupError):
            # If running from a tarball (e.g. release tarball) downloaded from github, we will not have the .git
            # folder available. Thus, setuptools_scm cannot determine the version in any way. Thus, use the
            # fallback_version from the pyproject.toml file (which doesn't exist when this is installed as a package,
            # but SHOULD exist when this is run from a downloaded tarball from git)

            # Pyproject.toml should be two levels up from this file
            import os.path
            pyproject_toml = "%s/../../pyproject.toml" % os.path.dirname(__file__)

            # Variables to track if we're in the right section and to store the fallback_version
            in_setuptools_scm_section = False
            fallback_version = None

            file = None
            try:
                file = open(pyproject_toml, 'r')
                # Open the file and parse it manually
                fallback_version = None
                with file:
                    for line in file:
                        stripped_line = line.strip()

                        # Check if we're entering the [tool.setuptools_scm] section
                        if stripped_line == "[tool.setuptools_scm]":
                            in_setuptools_scm_section = True
                        elif stripped_line.startswith("[") and in_setuptools_scm_section:
                            # We've reached a new section, so stop searching
                            break

                        # If we're in the right section, look for the fallback_version key
                        if in_setuptools_scm_section and stripped_line.startswith("fallback_version"):
                            # Extract the value after the '=' sign and strip any surrounding quotes or whitespace
                            fallback_version = stripped_line.split('=', 1)[1].strip().strip('"').strip("'")
                            break
                # Account for the possibility that we failed to extract the fallback_version field from pyproject.toml
                if fallback_version:
                    __version__ = fallback_version
                else:
                    msg = "fallback_version not found in file %s" % pyproject_toml
                    msg += " when trying the get the EESSI test suite version. This should never happen."
                    msg += " Please report an issue on Github, including information on how you installed"
                    msg += " the EESSI test suite."
                    print(msg)
            except FileNotFoundError:
                msg = "File %s not found when trying to extract the EESSI test suite version from" % pyproject_toml
                msg += " pyproject.toml. This should never happen. Please report an issue on GitHub,"
                msg += " including information on how you installed the EESSI test suite."
                print(msg)
            except Exception as e:
                print("When trying to open file %s, an exception was raised: %s." % (pyproject_toml, e))

# One of the above three methods to get __version__ defined SHOULD work in any situation.
# It's considered a bug you reach this point without having a __version__ set
if not __version__:
    msg = "__version__ should have been defined by now, but it is not."
    msg += " This is considered a bug, please report it in an issue on Github for the"
    msg += " EESSI test suite."
    raise ValueError(msg)



"""
This module tests the binary 'mw' in available modules containing substring 'MetalWalls'.
Test input files are defined in MetalWalls's repo under hackathonGPU/benchmark*,
see https://github.com/reframe-hpc/reframe/blob/develop/hpctestlib/sciapps/qespresso/benchmarks.py

ReFrame terminology:

"pipeline stages":
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#pipeline-hooks

"test parameter": a list of values, which will generate different test variants.
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.parameter

"test variant": a version of a test with a specific value for each test parameter
https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#test-variants

"concrete test cases": all test combinations that will actually run:
- test variants
- valid system:partition+programming environment combinations
https://reframe-hpc.readthedocs.io/en/stable/tutorial_deps.html#listing-dependencies

Tests can be filtered by name, tag, programming environment, system, partition, or maintainer,
see https://reframe-hpc.readthedocs.io/en/stable/manpage.html#test-filtering

Hooks acting on all possible test combinations (before filtering) are called after the 'init' stage.
Hooks acting on concrete test cases (after filtering) are called after the 'setup' stage.

See also https://reframe-hpc.readthedocs.io/en/stable/pipeline.html
"""
import reframe as rfm
from hpctestlib.sciapps.metalwalls.benchmarks import MetalWallsCheck
from reframe.core.builtins import run_after
from reframe.core.parameters import TestParam as parameter

from eessi.testsuite import hooks
from eessi.testsuite.constants import (COMPUTE_UNIT, CPU, DEVICE_TYPES, GPU, CI,
                                       SCALES, TAGS)
from eessi.testsuite.utils import find_modules, log


@rfm.simple_test
class EESSI_MetalWalls_MW(MetalWallsCheck):
    """MetalWalls benchmark tests.

    `MetalWalls <https://gitlab.com/ampere2/metalwalls>`__ """

    scale = parameter(SCALES.keys())

    valid_systems = ['*']
    valid_prog_environs = ['default']
    time_limit = '60m'
    # input files are downloaded
    readonly_files = ['']

    module_name = parameter(find_modules('MetalWalls'))
    # For now, MetalWalls is being build for CPU targets only
    # compute_device = parameter([DEVICE_TYPES[CPU], DEVICE_TYPES[GPU]])
    compute_device = parameter([DEVICE_TYPES[CPU], ])

    @run_after('init')
    def run_after_init(self):
        """Hooks to run after the init phase"""

        # Filter on which scales are supported by the partitions defined in the ReFrame configuration
        hooks.filter_supported_scales(self)

        # Make sure that GPU tests run in partitions that support running on a GPU,
        # and that CPU-only tests run in partitions that support running CPU-only.
        # Also support setting valid_systems on the cmd line.
        hooks.filter_valid_systems_by_device_type(self, required_device_type=self.compute_device)

        # Support selecting modules on the cmd line.
        hooks.set_modules(self)

        # Support selecting scales on the cmd line via tags.
        hooks.set_tag_scale(self)

    @run_after('init')
    def set_tag_ci(self):
        """Set tag CI on smallest benchmark, so it can be selected on the cmd line via --tag CI"""
        if self.benchmark_info[0] == 'hackathonGPU/benchmark':
            self.tags.add(TAGS[CI])
            log(f'tags set to {self.tags}')

    @run_after('init')
    def set_increased_walltime(self):
        """Increase the amount of time for the largest benchmark, when running with few cores."""
        # List of benchmarks that require more time to run
        large_benchmarks = ['hackathonGPU/benchmark2']
        if self.num_tasks <= 4 and self.benchmark_info[0] in large_benchmarks:
            self.time_limit = '120m'

    @run_after('setup')
    def run_after_setup(self):
        """Hooks to run after the setup phase"""

        # Calculate default requested resources based on the scale:
        # 1 task per CPU for CPU-only tests, 1 task per GPU for GPU tests.
        # Also support setting the resources on the cmd line.
        if self.compute_device == DEVICE_TYPES[GPU]:
            hooks.assign_tasks_per_compute_unit(test=self, compute_unit=COMPUTE_UNIT[GPU])
        else:
            hooks.assign_tasks_per_compute_unit(test=self, compute_unit=COMPUTE_UNIT[CPU])

    @run_after('setup')
    def set_binding(self):
        """Set binding to compact to improve performance reproducibility."""
        hooks.set_compact_process_binding(self)

    @run_after('setup')
    def request_mem(self):
        """Request memory per node based on the benchmark."""
        mem_per_task = 0.4
        if self.benchmark_info[0] == 'hackathonGPU/benchmark5':
            mem_per_task = 1.2
        memory_required = self.num_tasks_per_node * mem_per_task + 2
        hooks.req_memory_per_node(test=self, app_mem_req=memory_required * 1024)

    @run_after('setup')
    def skip_max_corecnt(self):
        """Skip tests if number of tasks per node exceeds maximum core count."""
        max_task_cnt = 256
        bench_name = self.benchmark_info[0]
        self.skip_if(
            self.num_tasks > max_task_cnt,
            f'Number of tasks {self.num_tasks} exceeds maximum task count {max_task_cnt} for {bench_name}'
        )

    @run_after('setup')
    def set_omp_num_threads(self):
        """
        Set number of OpenMP threads via OMP_NUM_THREADS.
        Set default number of OpenMP threads equal to number of CPUs per task.
        """
        hooks.set_omp_num_threads(self)



#!/usr/bin/env python
"""
MPI_Reduce on MPI rank. This should result in a total of (size * (size - 1) / 2),
where size is the total number of ranks.
Prints the total number of ranks, the sum of all ranks, and the time elapsed for the reduction."
"""

import argparse
import time

from mpi4py import MPI

parser = argparse.ArgumentParser(description='mpi4py reduction benchmark',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--n_warmup', type=int, default=100,
                    help='Number of warmup iterations')
parser.add_argument('--n_iter', type=int, default=1000,
                    help='Number of benchmark iterations')
args = parser.parse_args()

n_warmup = args.n_warmup
n_iter = args.n_iter

size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()
name = MPI.Get_processor_name()

# Warmup
t0 = time.time()
for i in range(n_warmup):
    total = MPI.COMM_WORLD.reduce(rank, op=MPI.SUM)

# Actual reduction, multiple iterations for accuracy of timing
t1 = time.time()
for i in range(n_iter):
    total = MPI.COMM_WORLD.reduce(rank, op=MPI.SUM)
t2 = time.time()
total_time = (t2 - t1) / n_iter

if rank == 0:
    print(f"Total ranks: {size}")
    print(f"Sum of all ranks: {total}")  # Should be (size * (size-1) / 2)
    print(f"Time elapsed: {total_time:.3e}")
